import dataclasses
import logging
from datetime import datetime, timezone
from typing import Optional

from pyhanko_certvalidator.context import (
    CertValidationPolicySpec,
    ValidationDataHandlers,
)
from pyhanko_certvalidator.errors import ValidationError
from pyhanko_certvalidator.ltv.errors import (
    PastValidatePrecheckFailure,
    TimeSlideFailure,
)
from pyhanko_certvalidator.ltv.time_slide import time_slide
from pyhanko_certvalidator.ltv.types import ValidationTimingInfo
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.policy_decl import (
    NO_REVOCATION,
    AcceptAllAlgorithms,
    CertRevTrustPolicy,
)
from pyhanko_certvalidator.validate import async_validate_path

__all__ = ['past_validate']

logger = logging.getLogger(__name__)


async def _past_validate_precheck(
    path: ValidationPath,
    validation_policy_spec: CertValidationPolicySpec,
):
    # The past validation algorithm requires us to run the "regular"
    # validation algorithm without regard for revocation and expiration
    # on a known-good time

    # Shell model: intersect the validity windows of all certs in the path
    certs = list(path.iter_certs(include_root=False))
    lower_bound = max(c.not_valid_before for c in certs)
    upper_bound = min(c.not_valid_after for c in certs)

    if lower_bound >= upper_bound:
        raise PastValidatePrecheckFailure(
            "The intersection of the validity periods of the certificates "
            "in the path is empty or degenerate."
        )

    ref_time = ValidationTimingInfo(
        validation_time=upper_bound,
        best_signature_time=upper_bound,
        point_in_time_validation=True,
    )

    validation_context = dataclasses.replace(
        validation_policy_spec,
        revinfo_policy=CertRevTrustPolicy(
            revocation_checking_policy=NO_REVOCATION
        ),
        algorithm_usage_policy=AcceptAllAlgorithms(),
    ).build_validation_context(timing_info=ref_time, handlers=None)

    try:
        await async_validate_path(
            validation_context,
            path,
            validation_policy_spec.pkix_validation_params,
        )
    except ValidationError as e:
        raise PastValidatePrecheckFailure(
            "Elementary path validation routine failed during pre-check "
            "for past point-in-time validation"
        ) from e


async def past_validate(
    path: ValidationPath,
    validation_policy_spec: CertValidationPolicySpec,
    validation_data_handlers: ValidationDataHandlers,
    init_control_time: Optional[datetime] = None,
    best_signature_time: Optional[datetime] = None,
) -> datetime:
    """
    Execute the ETSI EN 319 102-1 past certificate validation algorithm
    against the given path (ETSI EN 319 102-1, § 5.6.2.1).

    Instead of merely evaluating X.509 validation constraints, the algorithm
    will perform a full point-in-time reevaluation of the path at the
    control time mandated by the specification. This implies that a caller
    implementing the past signature validation algorithm no longer needs to
    explicitly reevaluate CA certificate revocation times and/or algorithm
    constraints based on POEs.

    .. warning::
        This is incubating internal API.

    :param path:
        The prospective validation path against which to execute the algorithm.
    :param validation_policy_spec:
        The validation policy specification.
    :param validation_data_handlers:
        The handlers used to manage collected certificates,revocation
        information and proof-of-existence records.
    :param init_control_time:
        Initial control time; defaults to the current time.
    :param best_signature_time:
        Usage time to use in freshness computations.
    :return:
        The control time returned by the time sliding algorithm.
        Informally, the last time at which the certificate was known to be
        valid.
    """

    await _past_validate_precheck(
        path,
        validation_policy_spec,
    )

    try:
        # time slide
        init_control_time = init_control_time or datetime.now(tz=timezone.utc)
        control_time = await time_slide(
            path,
            init_control_time=init_control_time,
            rev_trust_policy=validation_policy_spec.revinfo_policy,
            algo_usage_policy=validation_policy_spec.algorithm_usage_policy,
            time_tolerance=validation_policy_spec.time_tolerance,
            revinfo_manager=validation_data_handlers.revinfo_manager,
        )
        logger.info(
            f"AdES time slide yields %s as the control time for path with "
            f"leaf {path.describe_leaf()}",
            control_time,
        )
    except ValidationError as e:
        raise TimeSlideFailure(
            f"Failed to get control time for point-in-time validation for path "
            f"with leaf {path.describe_leaf()}"
        ) from e

    ref_time = ValidationTimingInfo(
        validation_time=control_time,
        best_signature_time=best_signature_time or control_time,
        point_in_time_validation=True,
    )

    # -> validate
    validation_context = validation_policy_spec.build_validation_context(
        timing_info=ref_time, handlers=validation_data_handlers
    )

    # Maintenance note:
    #  Doing a full point-in-time re-validation of the path is much more
    #  heavy-handed than what the AdES spec requires. We really only have to
    #  evaluate the chain constraints here.
    #  However, the past signature validation algorithm needs information about
    #  revocations up the chain and algorithm usage for _all_ operations in
    #  the validation process which is hard to pass on given the current
    #  architecture of certvalidator. Reevaluating with a time in the past
    #  is easier, and the POE enforcement is the same either way.

    await async_validate_path(
        validation_context,
        path,
        parameters=validation_policy_spec.pkix_validation_params,
    )

    return control_time



# coding: utf-8

import pytest

from pyhanko_certvalidator.fetchers.requests_fetchers import (
    RequestsCertificateFetcher,
)
from pyhanko_certvalidator.registry import (
    CertificateRegistry,
    PathBuilder,
    SimpleTrustManager,
)

from .common import load_cert_object


def test_build_paths_custom_ca_certs():
    cert = load_cert_object('mozilla.org.crt')
    other_certs = [load_cert_object('digicert-sha2-secure-server-ca.crt')]

    builder = PathBuilder(
        trust_manager=SimpleTrustManager.build(trust_roots=other_certs),
        registry=CertificateRegistry.build(certs=other_certs),
    )
    paths = builder.build_paths(cert)
    assert 1 == len(paths)

    path = paths[0]
    assert 2 == len(path)
    assert [item.subject.sha1 for item in path] == [
        b"\x10_\xa6z\x80\x08\x9d\xb5'\x9f5\xce\x83\x0bC\x88\x9e\xa3\xc7\r",
        b'I\xac\x03\xf8\xf3Km\xca)V)\xf2I\x9a\x98\xbe\x98\xdc.\x81',
    ]


@pytest.mark.parametrize(
    'domain',
    [
        "google.com",
        "www.cnn.com",
        "microsoft.com",
        "southwest.com",
        "xuite.net",
        "icpedu.rnp.br",
    ],
)
@pytest.mark.asyncio
async def test_basic_certificate_validator_tls_aia(domain):
    # google.com    -> application/pkix-cert
    # www.cnn.com   -> application/x-x509-ca-cert
    # microsoft.com -> application/octet-stream (DER)
    # southwest.com -> application/pkcs7-mime
    # xuite.net     -> application/x-pkcs7-certificates
    # icpedu.rnp.br -> binary/octet-stream (PEM, PKCS#7)

    icpedu_root = load_cert_object('testing-aia', 'root-icpedu.rnp.br')
    trust_manager = SimpleTrustManager.build(
        extra_trust_roots=[icpedu_root],
    )
    cert = load_cert_object('testing-aia', domain)
    registry = CertificateRegistry.build(
        certs=(cert,),
        cert_fetcher=RequestsCertificateFetcher(per_request_timeout=30),
    )
    builder = PathBuilder(trust_manager=trust_manager, registry=registry)
    paths = await builder.async_build_paths(end_entity_cert=cert)
    assert len(paths) >= 1



import asyncio
from datetime import datetime, timedelta
from typing import Iterable, List, Optional, Set, Tuple

from asn1crypto import algos, keys, x509

from pyhanko_certvalidator._state import ValProcState
from pyhanko_certvalidator.errors import (
    DisallowedAlgorithmError,
    InsufficientPOEError,
    InsufficientRevinfoError,
    RevokedError,
)
from pyhanko_certvalidator.ltv.types import (
    ValidationTimingInfo,
    ValidationTimingParams,
)
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.policy_decl import (
    AlgorithmUsagePolicy,
    CertRevTrustPolicy,
    RevocationCheckingRule,
)
from pyhanko_certvalidator.revinfo.archival import RevinfoContainer
from pyhanko_certvalidator.revinfo.manager import RevinfoManager
from pyhanko_certvalidator.revinfo.validate_crl import (
    CRLOfInterest,
    _check_cert_on_crl_and_delta,
    _CRLErrs,
    collect_relevant_crls_with_paths,
)
from pyhanko_certvalidator.revinfo.validate_ocsp import (
    OCSPResponseOfInterest,
    _check_ocsp_status,
    collect_relevant_responses_with_paths,
)
from pyhanko_certvalidator.util import ConsList

__all__ = ['time_slide', 'ades_gather_prima_facie_revinfo']


async def ades_gather_prima_facie_revinfo(
    path: ValidationPath,
    revinfo_manager: RevinfoManager,
    control_time: datetime,
    revocation_checking_rule: RevocationCheckingRule,
) -> Tuple[List[CRLOfInterest], List[OCSPResponseOfInterest]]:
    """
    Gather potentially relevant revocation information for the leaf
    certificate of a candidate validation path.
    Only the scope of the revocation information will be checked, no
    detailed validation will occur.

    :param path:
        The candidate validation path.
    :param revinfo_manager:
        The revocation info manager.
    :param control_time:
        The time horizon that serves as a relevance cutoff.
    :param revocation_checking_rule:
        Revocation info rule controlling which kind(s) of revocation
        information will be fetched.
    :return:
        A 2-element tuple containing a list of the fetched CRLs and
        OCSP responses, respectively.
    """

    cert = path.leaf
    if revocation_checking_rule.ocsp_relevant:
        ocsp_result = await collect_relevant_responses_with_paths(
            cert, path, revinfo_manager, control_time
        )
        ocsps = ocsp_result.responses
    else:
        ocsps = []

    if revocation_checking_rule.crl_relevant:
        crl_result = await collect_relevant_crls_with_paths(
            cert, path, revinfo_manager, control_time
        )
        crls = crl_result.crls
    else:
        crls = []
    return crls, ocsps


def _tails(path: ValidationPath):
    cur_path = path
    yield cur_path, True
    while cur_path.pkix_len > 1:
        cur_path = cur_path.copy_and_drop_leaf()
        yield cur_path, False


def _apply_algo_policy(
    algo_policy: AlgorithmUsagePolicy,
    algo_used: algos.SignedDigestAlgorithm,
    control_time: datetime,
    public_key: keys.PublicKeyInfo,
    val_proc_state: ValProcState,
):
    sig_constraint = algo_policy.signature_algorithm_allowed(
        algo_used, control_time, public_key
    )
    algo_name = algo_used['algorithm'].native
    if not sig_constraint.allowed:
        if sig_constraint.not_allowed_after:
            # rewind the clock up until the point where the algorithm
            # was actually permissible
            control_time = min(control_time, sig_constraint.not_allowed_after)
        else:
            msg = (
                f"Algorithm {algo_name} is banned outright without "
                f"time constraints."
            )
            if sig_constraint.failure_reason is not None:
                msg += f" Reason: {sig_constraint.failure_reason}"
            raise DisallowedAlgorithmError.from_state(
                msg,
                val_proc_state,
                banned_since=None,
            )
    return control_time


def _update_control_time_for_unrevoked(
    control_time: datetime,
    revinfo_container: RevinfoContainer,
    rev_trust_policy: CertRevTrustPolicy,
    time_tolerance: timedelta,
):
    # if the cert is not on the list, we need the freshness check
    usability = revinfo_container.usable_at(
        rev_trust_policy,
        ValidationTimingParams(
            timing_info=ValidationTimingInfo(
                validation_time=control_time,
                best_signature_time=control_time,
                point_in_time_validation=True,
            ),
            time_tolerance=time_tolerance,
        ),
    )
    issuance_date = revinfo_container.issuance_date
    if not usability.rating.usable_ades:
        # set the control time to the issuance date / last usable date
        # (note: the TOO_NEW check is to prevent problems
        #  with freshness policies involving cooldown periods,
        #  which aren't really supported in the time sliding
        #  algorithm, but hey)
        # NOTE: the spec mandates using the issuance date here, but I believe
        # that's wrong: the last date at which the revinfo is still considered
        # fresh should be used instead. This distinction matters, since
        # (especially when CRLs are used) the issuance date of the revinfo
        # is often before the signature time.
        cutoff_date = usability.last_usable_at or issuance_date
        if cutoff_date is not None:
            control_time = min(cutoff_date, control_time)
    return control_time


def _update_control_time(
    revoked_date: Optional[datetime],
    control_time: datetime,
    revinfo_container: RevinfoContainer,
    algo_policy: Optional[AlgorithmUsagePolicy],
    issuer_public_key: keys.PublicKeyInfo,
    val_proc_state: ValProcState,
):
    if revoked_date:
        # this means we have to update control_time
        control_time = min(revoked_date, control_time)
    algo_used = revinfo_container.revinfo_sig_mechanism_used
    if algo_policy is not None and algo_used is not None:
        control_time = _apply_algo_policy(
            algo_policy,
            algo_used,
            control_time,
            issuer_public_key,
            val_proc_state,
        )
    return control_time


async def _time_slide(
    path: ValidationPath,
    init_control_time: datetime,
    revinfo_manager: RevinfoManager,
    rev_trust_policy: CertRevTrustPolicy,
    algo_usage_policy: Optional[AlgorithmUsagePolicy],
    # TODO use policy objects
    time_tolerance: timedelta,
    cert_stack: ConsList[bytes],
    path_stack: ConsList[ValidationPath],
) -> datetime:
    control_time = init_control_time
    checking_policy = rev_trust_policy.revocation_checking_policy

    # For zero-length paths, there is nothing to check
    if path.pkix_len == 0:
        return init_control_time

    # The ETSI algorithm requires us to collect revinfo for each
    # cert in the path, starting with the first (after the root).
    # Since our revinfo collection methods require paths instead of individual
    # certs, we instead loop over partial paths
    partial_paths = list(reversed(list(_tails(path))))
    poe_manager = revinfo_manager.poe_manager
    for current_path, is_ee in partial_paths:
        crls, ocsps = await ades_gather_prima_facie_revinfo(
            current_path,
            revinfo_manager=revinfo_manager,
            control_time=control_time,
            revocation_checking_rule=(
                checking_policy.ee_certificate_rule
                if is_ee
                else checking_policy.intermediate_ca_cert_rule
            ),
        )
        cert = current_path.leaf
        new_cert_stack = cert_stack.cons(cert.dump())
        new_path_stack = path_stack.cons(path)

        proc_state = ValProcState(cert_path_stack=new_path_stack)

        if poe_manager[cert] > control_time:
            raise InsufficientPOEError.from_state(
                f"No proof of existence available for certificate "
                f"{cert.subject.human_friendly} at control time "
                f"{control_time.isoformat()}.",
                proc_state,
            )
        if not crls and not ocsps:
            if isinstance(cert, x509.Certificate):
                ident = cert.subject.human_friendly
            else:
                ident = "attribute certificate"

            # don't raise an error for revo-exempt certs (OCSP responders)
            if cert.ocsp_no_check_value is None:
                raise InsufficientRevinfoError.from_state(
                    f"No revocation info from before {control_time.isoformat()}"
                    f" found for certificate {ident}.",
                    proc_state,
                )

        once_revoked = False
        most_recent_crl = None
        # We always take the chain of trust of a CRL/OCSP response
        # at face value
        for crl_of_interest in crls:
            # skip CRLs that are no longer relevant
            issued = crl_of_interest.crl.issuance_date
            if (
                not issued
                or issued > control_time
                or poe_manager[crl_of_interest.crl] > control_time
            ):
                continue
            sub_paths = crl_of_interest.prov_paths

            # recurse into the paths associated with the CRL and adjust
            # the control time accordingly
            # don't bother checking issuers that already appear
            # in the chain of trust that we're currently looking into
            sub_path_skip_list: Set[bytes] = set(new_cert_stack) | set(
                cert.dump() for cert in current_path
            )
            sub_path_control_times = await asyncio.gather(
                *(
                    _time_slide(
                        crl_path.path,
                        control_time,
                        revinfo_manager,
                        rev_trust_policy,
                        algo_usage_policy,
                        time_tolerance,
                        cert_stack=new_cert_stack,
                        path_stack=new_path_stack,
                    )
                    for crl_path in sub_paths
                    if (
                        crl_path.path.leaf
                        and crl_path.path.leaf.dump() not in sub_path_skip_list
                    )
                )
            )
            control_time = min([control_time, *sub_path_control_times])

            for candidate_crl_path in sub_paths:
                revoked_date, revoked_reason = _check_cert_on_crl_and_delta(
                    crl_issuer=candidate_crl_path.path.leaf,
                    cert=cert,
                    certificate_list_cont=crl_of_interest.crl,
                    delta_certificate_list_cont=candidate_crl_path.delta,
                    errs=_CRLErrs(),
                )
                crl_iss_cert = candidate_crl_path.path.leaf
                assert isinstance(crl_iss_cert, x509.Certificate)

                once_revoked |= revoked_date is not None

                crl_container = crl_of_interest.crl
                if (
                    most_recent_crl is None
                    or most_recent_crl.issuance_date
                    < crl_container.issuance_date
                ):
                    most_recent_crl = crl_container
                control_time = _update_control_time(
                    revoked_date,
                    control_time,
                    revinfo_container=crl_container,
                    algo_policy=algo_usage_policy,
                    issuer_public_key=crl_iss_cert.public_key,
                    val_proc_state=proc_state,
                )

        most_recent_ocsp = None
        for ocsp_of_interest in ocsps:
            ocsp_container = ocsp_of_interest.ocsp_response
            issued = ocsp_container.issuance_date
            if (
                not issued
                or issued > control_time
                or poe_manager[ocsp_of_interest.ocsp_response] > control_time
            ):
                continue

            control_time = await _time_slide(
                ocsp_of_interest.prov_path,
                control_time,
                revinfo_manager,
                rev_trust_policy,
                algo_usage_policy,
                time_tolerance,
                cert_stack=new_cert_stack,
                path_stack=new_path_stack,
            )
            try:
                _check_ocsp_status(
                    ocsp_response=ocsp_container,
                    proc_state=ValProcState(cert_path_stack=new_path_stack),
                    control_time=control_time,
                )
                revoked_date = None
            except RevokedError as e:
                revoked_date = e.revocation_dt

            once_revoked |= revoked_date is not None
            ocsp_iss_cert = ocsp_of_interest.prov_path.leaf
            assert isinstance(ocsp_iss_cert, x509.Certificate)
            if (
                most_recent_ocsp is None
                or most_recent_ocsp.issuance_date < issued
            ):
                most_recent_ocsp = ocsp_container
            control_time = _update_control_time(
                revoked_date,
                control_time,
                revinfo_container=ocsp_container,
                algo_policy=algo_usage_policy,
                issuer_public_key=ocsp_iss_cert.public_key,
                val_proc_state=proc_state,
            )
        # check the algorithm constraints for the certificate itself
        if algo_usage_policy is not None:
            leaf_ca = list(current_path.iter_authorities())[-1]
            control_time = _apply_algo_policy(
                algo_usage_policy,
                cert['signature_algorithm'],
                control_time,
                leaf_ca.public_key,
                val_proc_state=proc_state,
            )

        # (c) if the certificate was not marked as revoked -> update
        # based on the freshness of the most recent piece of revinfo
        if not once_revoked:
            revinfo_items: Iterable[RevinfoContainer] = [
                x for x in (most_recent_ocsp, most_recent_crl) if x is not None
            ]
            most_recent_revinfo = max(
                revinfo_items,
                key=lambda x: x.issuance_date or control_time,
                default=None,
            )
            if most_recent_revinfo is not None:
                control_time = _update_control_time_for_unrevoked(
                    control_time=control_time,
                    revinfo_container=most_recent_revinfo,
                    rev_trust_policy=rev_trust_policy,
                    time_tolerance=time_tolerance,
                )

    return control_time


async def time_slide(
    path: ValidationPath,
    init_control_time: datetime,
    revinfo_manager: RevinfoManager,
    rev_trust_policy: CertRevTrustPolicy,
    algo_usage_policy: Optional[AlgorithmUsagePolicy],
    time_tolerance: timedelta,
) -> datetime:
    """
    Execute the ETSI EN 319 102-1 time slide algorithm against the given path.

    .. warning::
        This is incubating internal API.

    .. note::
        This implementation will also attempt to take into account chains of
        trust of indirect CRLs. This is not a requirement of the specification,
        but also somewhat unlikely to arise in practice in cases where AdES
        compliance actually matters.

    :param path:
        The prospective validation path against which to execute the time slide
        algorithm.
    :param init_control_time:
        The initial control time, typically the current time.
    :param revinfo_manager:
        The revocation info manager.
    :param rev_trust_policy:
        The trust policy for revocation information.
    :param algo_usage_policy:
        The algorithm usage policy.
    :param time_tolerance:
        The tolerance to apply when evaluating time-related constraints.
    :return:
        The resulting control time.
    """
    return await _time_slide(
        path,
        init_control_time,
        revinfo_manager,
        rev_trust_policy,
        algo_usage_policy,
        time_tolerance,
        cert_stack=ConsList.empty(),
        path_stack=ConsList.empty(),
    )



# coding: utf-8

import os

import pytest
from asn1crypto import pem, x509

from pyhanko_certvalidator.context import ValidationContext
from pyhanko_certvalidator.errors import OCSPFetchError
from pyhanko_certvalidator.fetchers import aiohttp_fetchers, requests_fetchers
from pyhanko_certvalidator.registry import (
    CertificateRegistry,
    PathBuilder,
    SimpleTrustManager,
)
from pyhanko_certvalidator.revinfo.validate_ocsp import verify_ocsp_response

from .common import load_cert_object
from .constants import TEST_REQUEST_TIMEOUT


@pytest.mark.asyncio
async def _test_with_fetchers(fetchers):
    intermediate = load_cert_object('digicert-g5-ecc-sha384-2021-ca1.crt')

    trust_roots = [load_cert_object(os.path.join('digicert-root-g5.crt'))]
    path_builder = PathBuilder(
        registry=CertificateRegistry.build(cert_fetcher=fetchers.cert_fetcher),
        trust_manager=SimpleTrustManager.build(trust_roots=trust_roots),
    )
    paths = await path_builder.async_build_paths(intermediate)
    path = paths[0]
    authority = path.find_issuing_authority(intermediate)

    ocsp_response = await fetchers.ocsp_fetcher.fetch(intermediate, authority)
    context = ValidationContext(
        trust_roots=trust_roots, ocsps=[ocsp_response], fetchers=fetchers
    )
    await verify_ocsp_response(intermediate, path, context)


async def _test_fetch_error(fetchers):
    # a cert that doesn't have any OCSP URLs will always throw an error
    cert_file = os.path.join('testing-ca-pss', 'interm.cert.pem')
    intermediate = load_cert_object(cert_file)

    root_file = os.path.join('testing-ca-pss', 'root.cert.pem')
    root = load_cert_object(root_file)

    path_builder = PathBuilder(
        registry=CertificateRegistry.build(cert_fetcher=fetchers.cert_fetcher),
        trust_manager=SimpleTrustManager.build(trust_roots=[root]),
    )
    paths = await path_builder.async_build_paths(intermediate)
    path = paths[0]
    authority = path.find_issuing_authority(intermediate)

    async def fetch_err():
        with pytest.raises(OCSPFetchError):
            await fetchers.ocsp_fetcher.fetch(intermediate, authority)

    # trigger this twice, to make sure we get an error for both jobs
    await fetch_err()
    await fetch_err()


@pytest.mark.asyncio
async def test_fetch_ocsp_aiohttp():
    fb = aiohttp_fetchers.AIOHttpFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    async with fb as fetchers:
        await _test_with_fetchers(fetchers)


@pytest.mark.asyncio
async def test_fetch_ocsp_err_aiohttp():
    fb = aiohttp_fetchers.AIOHttpFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    async with fb as fetchers:
        await _test_fetch_error(fetchers)


@pytest.mark.asyncio
async def test_fetch_ocsp_requests():
    fb = requests_fetchers.RequestsFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    fetchers = fb.get_fetchers()
    await _test_with_fetchers(fetchers)


@pytest.mark.asyncio
async def test_fetch_ocsp_err_requests():
    fb = requests_fetchers.RequestsFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    fetchers = fb.get_fetchers()
    await _test_fetch_error(fetchers)



import logging
from typing import Iterable, Union

import aiohttp
from asn1crypto import cms, ocsp, x509

from ... import errors
from ...authority import Authority
from ...util import get_ocsp_urls, issuer_serial
from ..api import OCSPFetcher
from ..common_utils import (
    format_ocsp_request,
    ocsp_job_get_earliest,
    process_ocsp_response_data,
)
from .util import AIOHttpMixin, LazySession

logger = logging.getLogger(__name__)


class AIOHttpOCSPFetcher(OCSPFetcher, AIOHttpMixin):
    def __init__(
        self,
        session: Union[aiohttp.ClientSession, LazySession],
        user_agent=None,
        per_request_timeout=10,
        certid_hash_algo='sha1',
        request_nonces=True,
    ):
        super().__init__(session, user_agent, per_request_timeout)
        if certid_hash_algo not in ('sha1', 'sha256'):
            raise ValueError(
                f'certid_hash_algo must be one of "sha1", "sha256", not '
                f'{repr(certid_hash_algo)}'
            )
        self.certid_hash_algo = certid_hash_algo
        self.request_nonces = request_nonces

    async def fetch(
        self,
        cert: Union[x509.Certificate, cms.AttributeCertificateV2],
        authority: Authority,
    ) -> ocsp.OCSPResponse:
        tag = (issuer_serial(cert), authority.hashable)
        if isinstance(cert, x509.Certificate):
            target = cert.subject.human_friendly
        else:
            # TODO log audit ID
            target = "attribute certificate"
        logger.info(f"About to queue OCSP fetch for {target}...")

        async def task():
            return await self._fetch(cert, authority)

        return await self._post_fetch_task(tag, task)

    async def _fetch(
        self,
        cert: Union[x509.Certificate, cms.AttributeCertificateV2],
        authority: Authority,
    ):
        ocsp_request = format_ocsp_request(
            cert,
            authority,
            certid_hash_algo=self.certid_hash_algo,
            request_nonces=self.request_nonces,
        )
        # Try the OCSP responders in arbitrary order, and process the responses
        # as they come in
        ocsp_urls = get_ocsp_urls(cert)
        if not ocsp_urls:
            raise errors.OCSPFetchError("No URLs to fetch OCSP responses from")

        if isinstance(cert, x509.Certificate):
            target = cert.subject.human_friendly
        else:
            # TODO log audit ID
            target = "attribute certificate"
        logger.info(
            f"Fetching OCSP status for {target} from url(s) "
            f"{';'.join(ocsp_urls)}..."
        )
        session = await self.get_session()
        fetch_jobs = (
            _grab_ocsp(
                ocsp_request,
                ocsp_url,
                user_agent=self.user_agent,
                session=session,
                timeout=self.per_request_timeout,
            )
            for ocsp_url in ocsp_urls
        )
        return await ocsp_job_get_earliest(fetch_jobs)

    def fetched_responses(self) -> Iterable[ocsp.OCSPResponse]:
        return self.get_results()

    def fetched_responses_for_cert(
        self, cert: x509.Certificate
    ) -> Iterable[ocsp.OCSPResponse]:
        target_is = issuer_serial(cert)
        return {
            resp
            for (subj_is, _), resp in self._iter_results()
            if subj_is == target_is
        }


async def _grab_ocsp(
    ocsp_request: ocsp.OCSPRequest,
    ocsp_url: str,
    *,
    user_agent,
    session: aiohttp.ClientSession,
    timeout,
):
    try:
        logger.info(f"Requesting OCSP response from {ocsp_url}...")
        headers = {
            'Accept': 'application/ocsp-response',
            'Content-Type': 'application/ocsp-request',
            'User-Agent': user_agent,
        }
        cl_timeout = aiohttp.ClientTimeout(total=timeout)
        async with session.post(
            url=ocsp_url,
            headers=headers,
            data=ocsp_request.dump(),
            raise_for_status=True,
            timeout=cl_timeout,
        ) as response:
            response_data = await response.read()
        return process_ocsp_response_data(
            response_data, ocsp_request=ocsp_request, ocsp_url=ocsp_url
        )
    except (aiohttp.ClientError, errors.OCSPValidationError) as e:
        raise errors.OCSPFetchError(
            f"Failed to fetch OCSP response from {ocsp_url}",
        ) from e



import datetime
import os

import pytest
from freezegun import freeze_time

from pyhanko_certvalidator.context import (
    CertValidationPolicySpec,
    ValidationDataHandlers,
)
from pyhanko_certvalidator.errors import (
    InsufficientPOEError,
    InsufficientRevinfoError,
)
from pyhanko_certvalidator.ltv.ades_past import past_validate
from pyhanko_certvalidator.ltv.errors import TimeSlideFailure
from pyhanko_certvalidator.ltv.poe import POEManager, POEType, digest_for_poe
from pyhanko_certvalidator.ltv.time_slide import time_slide
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.policy_decl import (
    CertRevTrustPolicy,
    FreshnessReqType,
    RevocationCheckingPolicy,
    RevocationCheckingRule,
)
from pyhanko_certvalidator.registry import (
    CertificateRegistry,
    SimpleTrustManager,
)
from pyhanko_certvalidator.revinfo.archival import CRLContainer, OCSPContainer
from pyhanko_certvalidator.revinfo.manager import RevinfoManager

from .common import load_cert_object, load_crl, load_ocsp_response, load_path

BASE_DIR = os.path.join('ades', 'time-slide')


def read_test_path(revoked_intermediate_ca=False) -> ValidationPath:
    return load_path(
        os.path.join(BASE_DIR, 'certs'),
        'root.crt',
        'interm-revoked.crt' if revoked_intermediate_ca else 'interm.crt',
        'alice.crt',
    )


def load_cert_registry(revoked_intermediate_ca=False) -> CertificateRegistry:
    cert_files = (
        'root.crt',
        'interm-revoked.crt' if revoked_intermediate_ca else 'interm.crt',
        'interm-ocsp.crt',
        'alice.crt',
    )
    reg = CertificateRegistry()
    for cert_file in cert_files:
        reg.register(load_cert_object(BASE_DIR, 'certs', cert_file))
    return reg


def now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


DEFAULT_REV_CHECK_POLICY = RevocationCheckingPolicy(
    ee_certificate_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
    intermediate_ca_cert_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
)
DEFAULT_TRUST_POLICY = CertRevTrustPolicy(
    revocation_checking_policy=DEFAULT_REV_CHECK_POLICY,
)

DEFAULT_TOLERANCE = datetime.timedelta(minutes=10)


@pytest.mark.asyncio
@freeze_time("2020-11-29T00:05:00+00:00")
async def test_time_slide_not_revoked():
    test_path = read_test_path()
    alice_ocsp = load_ocsp_response(BASE_DIR, 'alice-2020-11-29.ors')
    root_crl = load_crl(BASE_DIR, 'root-2020-11-29.crl')

    revinfo_manager = RevinfoManager(
        certificate_registry=load_cert_registry(),
        poe_manager=POEManager(),
        crls=[CRLContainer(root_crl)],
        ocsps=[OCSPContainer(alice_ocsp)],
    )
    control_time = await time_slide(
        test_path,
        init_control_time=now(),
        revinfo_manager=revinfo_manager,
        rev_trust_policy=DEFAULT_TRUST_POLICY,
        algo_usage_policy=None,
        time_tolerance=DEFAULT_TOLERANCE,
    )
    assert control_time == now()


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_time_slide_revocation_ocsp():
    test_path = read_test_path()
    alice_ocsp = load_ocsp_response(BASE_DIR, 'alice-2020-12-10.ors')
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')

    revinfo_manager = RevinfoManager(
        certificate_registry=load_cert_registry(),
        poe_manager=POEManager(),
        crls=[CRLContainer(root_crl)],
        ocsps=[OCSPContainer(alice_ocsp)],
    )
    control_time = await time_slide(
        test_path,
        init_control_time=now(),
        revinfo_manager=revinfo_manager,
        rev_trust_policy=DEFAULT_TRUST_POLICY,
        algo_usage_policy=None,
        time_tolerance=DEFAULT_TOLERANCE,
    )
    assert control_time == datetime.datetime(
        2020, 12, 1, tzinfo=datetime.timezone.utc
    )


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_time_slide_revocation_crl():
    test_path = read_test_path()
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')
    interm_crl = load_crl(BASE_DIR, 'interm-2020-12-10.crl')

    revinfo_manager = RevinfoManager(
        certificate_registry=load_cert_registry(),
        poe_manager=POEManager(),
        crls=[CRLContainer(root_crl), CRLContainer(interm_crl)],
        ocsps=[],
    )
    control_time = await time_slide(
        test_path,
        init_control_time=now(),
        revinfo_manager=revinfo_manager,
        rev_trust_policy=DEFAULT_TRUST_POLICY,
        algo_usage_policy=None,
        time_tolerance=DEFAULT_TOLERANCE,
    )
    assert control_time == datetime.datetime(
        2020, 12, 1, tzinfo=datetime.timezone.utc
    )


VERY_LENIENT_FRESHNESS = CertRevTrustPolicy(
    revocation_checking_policy=DEFAULT_REV_CHECK_POLICY,
    freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
    freshness=datetime.timedelta(days=100),
)


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_time_slide_revoked_intermediate():
    test_path = read_test_path(revoked_intermediate_ca=True)
    # the intermediate cert is listed as revoked on this CRL
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')
    # this CRL would be valid long enough to serve as non-revocation
    # evidence for the 'alice' cert
    # We set a ridiculous freshness window to ensure it's covered.
    poe_manager = POEManager()
    interm_crl = load_crl(BASE_DIR, 'interm-2020-11-29.crl')
    poe_date = datetime.datetime(2020, 11, 30, tzinfo=datetime.timezone.utc)
    poe_manager.register(test_path.leaf, dt=poe_date, poe_type=POEType.PROVIDED)
    # ...make sure to include some POE prior to the revocation date of the
    # intermediate cert
    poe_manager.register(
        CRLContainer(interm_crl), dt=poe_date, poe_type=POEType.PROVIDED
    )

    revinfo_manager = RevinfoManager(
        certificate_registry=load_cert_registry(revoked_intermediate_ca=True),
        poe_manager=poe_manager,
        crls=[CRLContainer(root_crl), CRLContainer(interm_crl)],
        ocsps=[],
    )
    control_time = await time_slide(
        test_path,
        init_control_time=now(),
        revinfo_manager=revinfo_manager,
        rev_trust_policy=VERY_LENIENT_FRESHNESS,
        algo_usage_policy=None,
        time_tolerance=DEFAULT_TOLERANCE,
    )
    assert control_time == datetime.datetime(
        2020, 12, 1, tzinfo=datetime.timezone.utc
    )


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_time_slide_revoked_intermediate_enforce_cert_poe():
    test_path = read_test_path(revoked_intermediate_ca=True)
    # the intermediate cert is listed as revoked on this CRL
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')
    poe_manager = POEManager()
    # No POE for the leaf cert at the control time
    # at which the intermediate cert was revoked => fail
    interm_crl = load_crl(BASE_DIR, 'interm-2020-11-29.crl')
    poe_date = datetime.datetime(2020, 11, 30, tzinfo=datetime.timezone.utc)
    poe_manager.register(
        CRLContainer(interm_crl), dt=poe_date, poe_type=POEType.PROVIDED
    )

    revinfo_manager = RevinfoManager(
        certificate_registry=load_cert_registry(revoked_intermediate_ca=True),
        poe_manager=poe_manager,
        crls=[CRLContainer(root_crl), CRLContainer(interm_crl)],
        ocsps=[],
    )
    with pytest.raises(InsufficientPOEError, match='for.*Alice'):
        await time_slide(
            test_path,
            init_control_time=now(),
            revinfo_manager=revinfo_manager,
            rev_trust_policy=VERY_LENIENT_FRESHNESS,
            algo_usage_policy=None,
            time_tolerance=DEFAULT_TOLERANCE,
        )


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_time_slide_revoked_intermediate_enforce_revinfo_poe():
    test_path = read_test_path(revoked_intermediate_ca=True)
    # the intermediate cert is listed as revoked on this CRL
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')
    poe_manager = POEManager()
    poe_date = datetime.datetime(2020, 11, 30, tzinfo=datetime.timezone.utc)
    poe_manager.register(test_path.leaf, dt=poe_date, poe_type=POEType.PROVIDED)
    # This CRL issued by the intermediate CA predates its revocation date
    # so without POE, it should be treated as no longer valid
    # => no revinfo for the leaf cert => can't finish
    interm_crl = load_crl(BASE_DIR, 'interm-2020-11-29.crl')

    revinfo_manager = RevinfoManager(
        certificate_registry=load_cert_registry(revoked_intermediate_ca=True),
        poe_manager=poe_manager,
        crls=[CRLContainer(root_crl), CRLContainer(interm_crl)],
        ocsps=[],
    )
    with pytest.raises(InsufficientRevinfoError, match='for.*Alice'):
        await time_slide(
            test_path,
            init_control_time=now(),
            revinfo_manager=revinfo_manager,
            rev_trust_policy=VERY_LENIENT_FRESHNESS,
            algo_usage_policy=None,
            time_tolerance=DEFAULT_TOLERANCE,
        )


VALIDATION_POLICY_SPEC = CertValidationPolicySpec(
    trust_manager=SimpleTrustManager.build(
        trust_roots=[load_cert_object(BASE_DIR, 'certs', 'root.crt')]
    ),
    revinfo_policy=DEFAULT_TRUST_POLICY,
)


@pytest.mark.asyncio
@freeze_time("2020-11-29T00:05:00+00:00")
async def test_point_in_time_validation_not_revoked():
    test_path = read_test_path()
    alice_ocsp = load_ocsp_response(BASE_DIR, 'alice-2020-11-29.ors')
    root_crl = load_crl(BASE_DIR, 'root-2020-11-29.crl')

    cert_registry = load_cert_registry()
    poe_manager = POEManager()
    revinfo_manager = RevinfoManager(
        certificate_registry=cert_registry,
        poe_manager=poe_manager,
        crls=[CRLContainer(root_crl)],
        ocsps=[OCSPContainer(alice_ocsp)],
    )

    last_valid_time = await past_validate(
        test_path,
        validation_policy_spec=VALIDATION_POLICY_SPEC,
        init_control_time=now(),
        validation_data_handlers=ValidationDataHandlers(
            revinfo_manager=revinfo_manager,
            poe_manager=poe_manager,
            cert_registry=cert_registry,
        ),
    )
    assert last_valid_time == now()


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_point_in_time_validation_revoked_intermediate():
    # Same scenario as the time slide test w/ revoked intermediate cert & PoE
    # in this module

    test_path = read_test_path(revoked_intermediate_ca=True)
    # the intermediate cert is listed as revoked on this CRL
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')
    # this CRL would be valid long enough to serve as non-revocation
    # evidence for the 'alice' cert
    # We set a ridiculous freshness window to ensure it's covered.
    poe_date = datetime.datetime(2020, 11, 30, tzinfo=datetime.timezone.utc)
    poe_manager = POEManager()
    poe_manager.register(test_path.leaf, dt=poe_date, poe_type=POEType.PROVIDED)
    interm_crl = load_crl(BASE_DIR, 'interm-2020-11-29.crl')
    # ...make sure to include some POE prior to the revocation date of the
    # intermediate cert
    poe_manager.register(
        CRLContainer(interm_crl), dt=poe_date, poe_type=POEType.PROVIDED
    )

    cert_registry = load_cert_registry(revoked_intermediate_ca=True)
    revinfo_manager = RevinfoManager(
        certificate_registry=cert_registry,
        poe_manager=poe_manager,
        crls=[CRLContainer(root_crl), CRLContainer(interm_crl)],
        ocsps=[],
    )
    last_valid_time = await past_validate(
        test_path,
        validation_policy_spec=VALIDATION_POLICY_SPEC,
        init_control_time=now(),
        validation_data_handlers=ValidationDataHandlers(
            revinfo_manager=revinfo_manager,
            poe_manager=poe_manager,
            cert_registry=cert_registry,
        ),
    )
    assert last_valid_time == datetime.datetime(
        2020, 12, 1, tzinfo=datetime.timezone.utc
    )


@pytest.mark.asyncio
@freeze_time("2020-12-10T00:05:00+00:00")
async def test_point_in_time_validation_revinfo_insufficient_poe():
    test_path = read_test_path(revoked_intermediate_ca=True)
    # the intermediate cert is listed as revoked on this CRL
    root_crl = load_crl(BASE_DIR, 'root-2020-12-10.crl')
    poe_manager = POEManager()

    cert_registry = load_cert_registry(revoked_intermediate_ca=True)
    revinfo_manager = RevinfoManager(
        certificate_registry=cert_registry,
        poe_manager=poe_manager,
        crls=[CRLContainer(root_crl)],
        ocsps=[],
    )
    with pytest.raises(TimeSlideFailure):
        await past_validate(
            test_path,
            validation_policy_spec=VALIDATION_POLICY_SPEC,
            init_control_time=now(),
            validation_data_handlers=ValidationDataHandlers(
                revinfo_manager=revinfo_manager,
                poe_manager=poe_manager,
                cert_registry=cert_registry,
            ),
        )


def test_poe_manager_read_cert():
    manager = POEManager()
    cert = load_cert_object(BASE_DIR, 'certs', 'root.crt')
    with freeze_time('2020-11-11'):
        manager.register(cert, poe_type=POEType.PROVIDED)
    assert manager[cert].date() == datetime.date(2020, 11, 11)


def test_poe_manager_cert_by_digest():
    manager = POEManager()
    cert = load_cert_object(BASE_DIR, 'certs', 'root.crt')
    with freeze_time('2020-11-11'):
        manager.register_by_digest(
            digest_for_poe(cert.dump()), poe_type=POEType.PROVIDED
        )
    assert manager[cert].date() == datetime.date(2020, 11, 11)


def test_poe_manager_read_bytes():
    manager = POEManager()
    msg = b'deadbeef'
    with freeze_time('2020-11-11'):
        manager.register(msg, poe_type=POEType.PROVIDED)
    assert manager[msg].date() == datetime.date(2020, 11, 11)


def test_poe_manager_read_bytes_by_digest():
    manager = POEManager()
    msg = b'deadbeef'
    with freeze_time('2020-11-11'):
        manager.register_by_digest(
            digest_for_poe(msg), poe_type=POEType.PROVIDED
        )
    assert manager[msg].date() == datetime.date(2020, 11, 11)



# coding: utf-8

import json
import os
from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterable, List, Optional, Type

import pytest
from asn1crypto import crl, ocsp, x509
from asn1crypto.util import timezone

from pyhanko_certvalidator import PKIXValidationParams
from pyhanko_certvalidator.authority import Authority, CertTrustAnchor
from pyhanko_certvalidator.context import ValidationContext
from pyhanko_certvalidator.errors import (
    CertificateFetchError,
    CRLFetchError,
    InsufficientRevinfoError,
    OCSPFetchError,
    OCSPValidationError,
    PathValidationError,
    RevokedError,
    StaleRevinfoError,
)
from pyhanko_certvalidator.fetchers import (
    CertificateFetcher,
    CRLFetcher,
    FetcherBackend,
    Fetchers,
    OCSPFetcher,
    aiohttp_fetchers,
    requests_fetchers,
)
from pyhanko_certvalidator.ltv.poe import POEManager
from pyhanko_certvalidator.path import QualifiedPolicy, ValidationPath
from pyhanko_certvalidator.policy_decl import (
    DisallowWeakAlgorithmsPolicy,
    NonRevokedStatusAssertion,
)
from pyhanko_certvalidator.registry import CertificateRegistry
from pyhanko_certvalidator.revinfo.manager import RevinfoManager
from pyhanko_certvalidator.validate import async_validate_path, validate_path

from .common import (
    FIXTURES_DIR,
    load_cert_object,
    load_nist_cert,
    load_nist_crl,
    load_openssl_ors,
)
from .constants import TEST_REQUEST_TIMEOUT


class MockOCSPFetcher(OCSPFetcher):
    def fetched_responses(self) -> Iterable[ocsp.OCSPResponse]:
        return ()

    def fetched_responses_for_cert(
        self, cert: x509.Certificate
    ) -> Iterable[ocsp.OCSPResponse]:
        return ()

    async def fetch(self, cert: x509.Certificate, authority: Authority):
        raise OCSPFetchError("No connection")


class MockOCSPFetcherWithValidationError(MockOCSPFetcher):
    async def fetch(self, cert: x509.Certificate, authority: Authority):
        raise OCSPValidationError("Something went wrong")


class MockCRLFetcher(CRLFetcher):
    def fetched_crls_for_cert(
        self, cert: x509.Certificate
    ) -> Iterable[crl.CertificateList]:
        return ()

    def fetched_crls(self) -> Iterable[crl.CertificateList]:
        return ()

    async def fetch(self, cert: x509.Certificate, *, use_deltas=None):
        raise CRLFetchError("No connection")


class MockCertFetcher(CertificateFetcher):
    def fetched_certs(self) -> Iterable[x509.Certificate]:
        return ()

    def fetch_cert_issuers(self, cert):
        return self

    def fetch_crl_issuers(self, certificate_list):
        return self

    def __aiter__(self):
        raise CertificateFetchError("No connection")


class MockFetcherBackend(FetcherBackend):
    def get_fetchers(self) -> Fetchers:
        return Fetchers(
            ocsp_fetcher=MockOCSPFetcher(),
            crl_fetcher=MockCRLFetcher(),
            cert_fetcher=MockCertFetcher(),
        )


class MockFetcherBackendWithValidationError(FetcherBackend):
    def get_fetchers(self) -> Fetchers:
        return Fetchers(
            ocsp_fetcher=MockOCSPFetcherWithValidationError(),
            crl_fetcher=MockCRLFetcher(),
            cert_fetcher=MockCertFetcher(),
        )


ERR_CLASSES = {
    cls.__name__: cls
    for cls in (
        PathValidationError,
        RevokedError,
        InsufficientRevinfoError,
        StaleRevinfoError,
    )
}


@dataclass(frozen=True)
class PKITSTestCaseErrorResult:
    err_class: Type[Exception]
    msg_regex: str


def test_revocation_mode_soft():
    cert = load_cert_object(
        'digicert-ecc-p384-root-g5-revoked-chain-demos-digicert-com.crt'
    )
    ca_certs = [load_cert_object('digicert-root-g5.crt')]
    other_certs = [
        load_cert_object('digicert-g5-ecc-sha384-2021-ca1.crt'),
    ]

    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        moment=datetime(2023, 1, 10, tzinfo=timezone.utc),
        allow_fetching=True,
        weak_hash_algos={'md2', 'md5'},
        fetcher_backend=MockFetcherBackend(),
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)

    validate_path(context, path)


def test_revocation_mode_soft_fail():
    cert = load_cert_object(
        'digicert-ecc-p384-root-g5-revoked-chain-demos-digicert-com.crt'
    )
    ca_certs = [load_cert_object('digicert-root-g5.crt')]
    other_certs = [
        load_cert_object('digicert-g5-ecc-sha384-2021-ca1.crt'),
    ]

    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        moment=datetime(2023, 1, 10, tzinfo=timezone.utc),
        allow_fetching=True,
        weak_hash_algos={'md2', 'md5'},
        fetcher_backend=MockFetcherBackendWithValidationError(),
    )
    paths = context.path_builder.build_paths(cert)
    path = paths[0]

    with pytest.raises(InsufficientRevinfoError, match="Something went wrong"):
        validate_path(context, path)


@pytest.mark.skip("annoying to maintain; replace with certomancer test")
def test_revocation_mode_hard():
    cert = load_cert_object(
        'digicert-ecc-p384-root-g5-revoked-chain-demos-digicert-com.crt'
    )
    ca_certs = [load_cert_object('digicert-root-g5.crt')]
    other_certs = [
        load_cert_object('digicert-g5-ecc-sha384-2021-ca1.crt'),
    ]

    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        allow_fetching=True,
        revocation_mode='hard-fail',
        weak_hash_algos={'md2', 'md5'},
        fetcher_backend=requests_fetchers.RequestsFetcherBackend(
            per_request_timeout=TEST_REQUEST_TIMEOUT
        ),
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)

    expected = (
        '(CRL|OCSP response) indicates the end-entity certificate was '
        'revoked at \\d\\d:\\d\\d:\\d\\d on \\d\\d\\d\\d-\\d\\d-\\d\\d'
        ', due to an unspecified reason'
    )
    with pytest.raises(RevokedError, match=expected):
        validate_path(context, path)


@pytest.mark.skip("annoying to maintain; replace with certomancer test")
@pytest.mark.asyncio
async def test_revocation_mode_hard_async():
    cert = load_cert_object(
        'digicert-ecc-p384-root-g5-revoked-chain-demos-digicert-com.crt'
    )
    ca_certs = [load_cert_object('digicert-root-g5.crt')]
    other_certs = [
        load_cert_object('digicert-g5-ecc-sha384-2021-ca1.crt'),
    ]
    fb = aiohttp_fetchers.AIOHttpFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    async with fb as fetchers:
        context = ValidationContext(
            trust_roots=ca_certs,
            other_certs=other_certs,
            allow_fetching=True,
            revocation_mode='hard-fail',
            weak_hash_algos={'md2', 'md5'},
            fetchers=fetchers,
        )
        paths = await context.path_builder.async_build_paths(cert)
        assert 1 == len(paths)
        path = paths[0]
        assert 3 == len(path)

        expected = (
            '(CRL|OCSP response) indicates the end-entity certificate was '
            'revoked at \\d\\d:\\d\\d:\\d\\d on \\d\\d\\d\\d-\\d\\d-\\d\\d'
            ', due to an unspecified reason'
        )
        with pytest.raises(RevokedError, match=expected):
            await async_validate_path(context, path)


@pytest.mark.skip("annoying to maintain; replace with certomancer test")
@pytest.mark.asyncio
async def test_revocation_mode_hard_aiohttp_autofetch():
    cert = load_cert_object(
        'digicert-ecc-p384-root-g5-revoked-chain-demos-digicert-com.crt'
    )
    ca_certs = [load_cert_object('digicert-root-g5.crt')]

    fb = aiohttp_fetchers.AIOHttpFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    async with fb as fetchers:
        context = ValidationContext(
            trust_roots=ca_certs,
            allow_fetching=True,
            revocation_mode='hard-fail',
            weak_hash_algos={'md2', 'md5'},
            fetchers=fetchers,
        )
        paths = await context.path_builder.async_build_paths(cert)
        assert 1 == len(paths)
        path = paths[0]
        assert 3 == len(path)

        expected = (
            '(CRL|OCSP response) indicates the end-entity certificate was '
            'revoked at \\d\\d:\\d\\d:\\d\\d on \\d\\d\\d\\d-\\d\\d-\\d\\d'
            ', due to an unspecified reason'
        )
        with pytest.raises(RevokedError, match=expected):
            await async_validate_path(context, path)


@pytest.mark.skip("annoying to maintain; replace with certomancer test")
@pytest.mark.asyncio
async def test_revocation_mode_hard_requests_autofetch():
    cert = load_cert_object(
        'digicert-ecc-p384-root-g5-revoked-chain-demos-digicert-com.crt'
    )
    ca_certs = [load_cert_object('digicert-root-g5.crt')]

    fb = requests_fetchers.RequestsFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    async with fb as fetchers:
        context = ValidationContext(
            trust_roots=ca_certs,
            allow_fetching=True,
            revocation_mode='hard-fail',
            weak_hash_algos={'md2', 'md5'},
            fetchers=fetchers,
        )
        paths = await context.path_builder.async_build_paths(cert)
        assert 1 == len(paths)
        path = paths[0]
        assert 3 == len(path)

        expected = (
            '(CRL|OCSP response) indicates the end-entity certificate was '
            'revoked at \\d\\d:\\d\\d:\\d\\d on \\d\\d\\d\\d-\\d\\d-\\d\\d'
            ', due to an unspecified reason'
        )
        with pytest.raises(RevokedError, match=expected):
            await async_validate_path(context, path)


def test_rsassa_pss():
    cert = load_cert_object('testing-ca-pss', 'signer1.cert.pem')
    ca_certs = [load_cert_object('testing-ca-pss', 'root.cert.pem')]
    other_certs = [load_cert_object('testing-ca-pss', 'interm.cert.pem')]
    moment = datetime(2021, 5, 3, tzinfo=timezone.utc)
    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        allow_fetching=False,
        moment=moment,
        revocation_mode='soft-fail',
        weak_hash_algos={'md2', 'md5'},
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)
    validate_path(context, path)


def test_rsassa_pss_exclusive():
    cert = load_cert_object('testing-ca-pss-exclusive', 'signer1.cert.pem')
    ca_certs = [load_cert_object('testing-ca-pss-exclusive', 'root.cert.pem')]
    other_certs = [
        load_cert_object('testing-ca-pss-exclusive', 'interm.cert.pem')
    ]
    moment = datetime(2021, 5, 3, tzinfo=timezone.utc)
    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        allow_fetching=False,
        moment=moment,
        revocation_mode='soft-fail',
        weak_hash_algos={'md2', 'md5'},
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)
    validate_path(context, path)


def test_ed25519():
    cert = load_cert_object('testing-ca-ed25519', 'signer.cert.pem')
    ca_certs = [load_cert_object('testing-ca-ed25519', 'root.cert.pem')]
    other_certs = [load_cert_object('testing-ca-ed25519', 'interm.cert.pem')]
    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        allow_fetching=False,
        revocation_mode='soft-fail',
        weak_hash_algos={'md2', 'md5'},
        moment=datetime(2020, 11, 1, tzinfo=timezone.utc),
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)
    validate_path(context, path)


def test_ed448():
    cert = load_cert_object('testing-ca-ed448', 'signer.cert.pem')
    ca_certs = [load_cert_object('testing-ca-ed448', 'root.cert.pem')]
    other_certs = [load_cert_object('testing-ca-ed448', 'interm.cert.pem')]
    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        allow_fetching=False,
        revocation_mode='soft-fail',
        weak_hash_algos={'md2', 'md5'},
        moment=datetime(2020, 11, 1, tzinfo=timezone.utc),
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)
    validate_path(context, path)


def test_assert_no_revinfo_needed_by_fiat():
    cert = load_cert_object('testing-ca-pss', 'signer1.cert.pem')
    ca_certs = [load_cert_object('testing-ca-pss', 'root.cert.pem')]
    other_certs = [load_cert_object('testing-ca-pss', 'interm.cert.pem')]
    moment = datetime(2021, 5, 3, tzinfo=timezone.utc)
    assertion = NonRevokedStatusAssertion(cert.sha256, moment)
    revinfo_manager = RevinfoManager(
        certificate_registry=CertificateRegistry.build(),
        poe_manager=POEManager(),
        crls=(),
        ocsps=(),
        assertions=(assertion,),
    )
    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        allow_fetching=False,
        moment=moment,
        revocation_mode='require',  # turn on strict revinfovalidation
        revinfo_manager=revinfo_manager,
    )
    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)
    validate_path(context, path)


def test_multitasking_ocsp():
    # regression test for case where the same responder ID (name + key ID)
    # is used in OCSP responses for different issuers in the same chain of
    # trust

    ors_dir = os.path.join(FIXTURES_DIR, 'multitasking-ocsp')
    with open(os.path.join(ors_dir, 'ocsp-resp-alice.der'), 'rb') as ocspin:
        ocsp_resp_alice = ocsp.OCSPResponse.load(ocspin.read())
    with open(os.path.join(ors_dir, 'ocsp-resp-interm.der'), 'rb') as ocspin:
        ocsp_resp_interm = ocsp.OCSPResponse.load(ocspin.read())
    vc = ValidationContext(
        trust_roots=[
            load_cert_object('multitasking-ocsp', 'root.cert.pem'),
        ],
        other_certs=[load_cert_object('multitasking-ocsp', 'interm.cert.pem')],
        revocation_mode='hard-fail',
        allow_fetching=False,
        ocsps=[ocsp_resp_interm, ocsp_resp_alice],
        moment=datetime(2021, 8, 19, 12, 20, 44, tzinfo=timezone.utc),
    )

    cert = load_cert_object('multitasking-ocsp', 'alice.cert.pem')
    paths = vc.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path = paths[0]
    assert 3 == len(path)
    validate_path(vc, path)


@dataclass(frozen=True)
class OCSPTestCase:
    name: str
    roots: List[x509.Certificate]
    cert: x509.Certificate
    ocsps: List[ocsp.OCSPResponse]
    path_len: int
    moment: datetime
    other_certs: List[x509.Certificate] = field(default_factory=list)
    expected_error: Optional[PKITSTestCaseErrorResult] = None

    @classmethod
    def from_json(cls, obj: dict):
        roots = [load_cert_object('openssl-ocsp', obj['root'])]
        kwargs = dict(
            name=obj['name'],
            cert=load_cert_object('openssl-ocsp', obj['cert']),
            path_len=int(obj['path_len']),
            moment=datetime.fromisoformat(obj['moment']),
            roots=roots,
        )
        kwargs['ocsps'] = [
            load_openssl_ors(filename) for filename in obj['ocsps']
        ]
        if 'other_certs' in obj:
            kwargs['other_certs'] = [
                load_cert_object('openssl-ocsp', filename)
                for filename in obj['other_certs']
            ]
        if 'error' in obj:
            kwargs['expected_error'] = PKITSTestCaseErrorResult(
                ERR_CLASSES[obj['error']['class']], obj['error']['msg_regex']
            )

        return OCSPTestCase(**kwargs)


def read_openssl_ocsp_test_params():
    data_path = os.path.join(FIXTURES_DIR, 'openssl-ocsp', 'openssl-ocsp.json')
    with open(data_path, 'r') as inf:
        cases = json.load(inf)
    return [OCSPTestCase.from_json(obj) for obj in cases]


@pytest.mark.parametrize(
    "test_case", read_openssl_ocsp_test_params(), ids=lambda case: case.name
)
def test_openssl_ocsp(test_case: OCSPTestCase):
    context = ValidationContext(
        trust_roots=test_case.roots,
        other_certs=test_case.other_certs,
        moment=test_case.moment,
        ocsps=test_case.ocsps,
        weak_hash_algos={'md2', 'md5'},
    )
    paths = context.path_builder.build_paths(test_case.cert)
    assert 1 == len(paths)
    path = paths[0]
    assert test_case.path_len == len(path)

    err = test_case.expected_error
    if err:
        with pytest.raises(err.err_class, match=err.msg_regex):
            validate_path(context, path)
    else:
        validate_path(context, path)


def parse_pkix_params(obj: dict):
    kwargs = {}
    if 'user_initial_policy_set' in obj:
        kwargs['user_initial_policy_set'] = frozenset(
            obj['user_initial_policy_set']
        )
    kwargs['initial_policy_mapping_inhibit'] = bool(
        obj.get('initial_policy_mapping_inhibit', False)
    )
    kwargs['initial_explicit_policy'] = bool(
        obj.get('initial_explicit_policy', False)
    )
    kwargs['initial_any_policy_inhibit'] = bool(
        obj.get('initial_any_policy_inhibit', False)
    )
    return PKIXValidationParams(**kwargs)


@dataclass(frozen=True)
class CannedTestInfo:
    test_id: int
    test_name: str

    def __str__(self):
        return f"{self.test_id} ({self.test_name})"


@dataclass(frozen=True)
class PKITSTestCase:
    test_info: CannedTestInfo
    cert: x509.Certificate
    roots: List[x509.Certificate]
    crls: List[crl.CertificateList]
    path_len: int
    path: Optional[ValidationPath] = None
    check_revocation: bool = True
    other_certs: List[x509.Certificate] = field(default_factory=list)
    expected_error: Optional[PKITSTestCaseErrorResult] = None
    pkix_params: Optional[PKIXValidationParams] = None

    @classmethod
    def from_json(cls, obj: dict):
        root = load_nist_cert('TrustAnchorRootCertificate.crt')
        crls = [load_nist_crl('TrustAnchorRootCRL.crl')]
        if 'crls' in obj:
            crls.extend(load_nist_crl(crl_path) for crl_path in obj['crls'])
        cert = load_nist_cert(obj['cert'])
        kwargs = dict(
            test_info=CannedTestInfo(
                test_id=int(obj['id']),
                test_name=obj['name'],
            ),
            cert=cert,
            path_len=int(obj['path_len']),
            check_revocation=bool(obj.get('revocation', True)),
            roots=[root],
            crls=crls,
        )

        kwargs['crls'] = crls
        if 'other_certs' in obj:
            kwargs['other_certs'] = [
                load_nist_cert(cert_path) for cert_path in obj['other_certs']
            ]
        if 'path_intermediates' in obj:
            # -> prebuild the path as indicated in the test spec
            kwargs['path'] = ValidationPath(
                trust_anchor=CertTrustAnchor(root),
                interm=(
                    load_nist_cert(cert_path)
                    for cert_path in obj['path_intermediates']
                ),
                leaf=cert,
            )
        if 'params' in obj:
            kwargs['pkix_params'] = parse_pkix_params(obj['params'])
        if 'error' in obj:
            kwargs['expected_error'] = PKITSTestCaseErrorResult(
                ERR_CLASSES[obj['error']['class']], obj['error']['msg_regex']
            )

        return PKITSTestCase(**kwargs)


def read_pkits_test_params():
    data_path = os.path.join(FIXTURES_DIR, 'nist_pkits', 'pkits.json')
    with open(data_path, 'r') as inf:
        cases = json.load(inf)
    return [PKITSTestCase.from_json(obj) for obj in cases]


@pytest.mark.parametrize(
    'test_case', read_pkits_test_params(), ids=lambda case: str(case.test_info)
)
def test_nist_pkits(test_case: PKITSTestCase):
    revocation_mode = "require" if test_case.check_revocation else "hard-fail"

    context = ValidationContext(
        trust_roots=test_case.roots,
        other_certs=test_case.other_certs,
        crls=test_case.crls,
        revocation_mode=revocation_mode,
        # adjust default algo policy to pass NIST tests
        algorithm_usage_policy=DisallowWeakAlgorithmsPolicy(
            weak_hash_algos={'md2', 'md5'}, dsa_key_size_threshold=1024
        ),
    )

    if test_case.path is None:
        paths = context.path_builder.build_paths(test_case.cert)
        assert 1 == len(paths)
        path: ValidationPath = paths[0]
    else:
        path = test_case.path

    assert test_case.path_len == len(path)

    err = test_case.expected_error
    params = test_case.pkix_params
    if err is not None:
        with pytest.raises(err.err_class, match=err.msg_regex):
            validate_path(context, path, parameters=params)
    else:
        validate_path(context, path, parameters=params)

        # sanity check
        if params is not None and params.user_initial_policy_set != {
            'any_policy'
        }:
            qps = path.qualified_policies()
            if qps is not None:
                for pol in qps:
                    assert (
                        pol.user_domain_policy_id
                        in params.user_initial_policy_set
                    )


@dataclass(frozen=True)
class PKITSUserNoticeTestCase:
    test_info: CannedTestInfo
    cert: x509.Certificate
    roots: List[x509.Certificate]
    crls: List[crl.CertificateList]
    notice: str
    other_certs: List[x509.Certificate] = field(default_factory=list)
    pkix_params: Optional[PKIXValidationParams] = None

    @classmethod
    def from_json(cls, obj: dict):
        roots = [load_nist_cert('TrustAnchorRootCertificate.crt')]
        crls = [load_nist_crl('TrustAnchorRootCRL.crl')]
        if 'crls' in obj:
            crls.extend(load_nist_crl(crl_path) for crl_path in obj['crls'])
        kwargs = dict(
            test_info=CannedTestInfo(
                test_id=int(obj['id']),
                test_name=obj['name'],
            ),
            cert=load_nist_cert(obj['cert']),
            roots=roots,
            crls=crls,
            notice=obj['notice'],
        )

        kwargs['crls'] = crls
        if 'other_certs' in obj:
            kwargs['other_certs'] = [
                load_nist_cert(cert_path) for cert_path in obj['other_certs']
            ]
        if 'params' in obj:
            kwargs['pkix_params'] = parse_pkix_params(obj['params'])

        return PKITSUserNoticeTestCase(**kwargs)


def read_pkits_user_notice_test_params():
    data_path = os.path.join(
        FIXTURES_DIR, 'nist_pkits', 'pkits-user-notice.json'
    )
    with open(data_path, 'r') as inf:
        cases = json.load(inf)
    return [PKITSUserNoticeTestCase.from_json(obj) for obj in cases]


@pytest.mark.parametrize(
    'test_case',
    read_pkits_user_notice_test_params(),
    ids=lambda case: str(case.test_info),
)
def test_nist_pkits_user_notice(test_case: PKITSUserNoticeTestCase):
    context = ValidationContext(
        trust_roots=test_case.roots,
        other_certs=test_case.other_certs,
        crls=test_case.crls,
        revocation_mode="require",
        weak_hash_algos={'md2', 'md5'},
    )

    paths = context.path_builder.build_paths(test_case.cert)
    assert 1 == len(paths)
    path: ValidationPath = paths[0]
    validate_path(context, path, parameters=test_case.pkix_params)

    qps = path.qualified_policies()
    assert 1 == len(qps)

    qp: QualifiedPolicy
    (qp,) = qps
    assert 1 == len(qp.qualifiers)
    (qual_obj,) = qp.qualifiers
    assert qual_obj['policy_qualifier_id'].native == 'user_notice'
    assert qual_obj['qualifier']['explicit_text'].native == test_case.notice


def test_408020_cps_pointer_qualifier_test20():
    cert = load_nist_cert('CPSPointerQualifierTest20EE.crt')
    ca_certs = [load_nist_cert('TrustAnchorRootCertificate.crt')]
    other_certs = [load_nist_cert('GoodCACert.crt')]
    crls = [
        load_nist_crl('GoodCACRL.crl'),
        load_nist_crl('TrustAnchorRootCRL.crl'),
    ]

    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        crls=crls,
        revocation_mode="require",
        weak_hash_algos={'md2', 'md5'},
    )

    paths = context.path_builder.build_paths(cert)
    assert 1 == len(paths)
    path: ValidationPath = paths[0]
    validate_path(context, path)

    qps = path.qualified_policies()
    assert 1 == len(qps)

    qp: QualifiedPolicy
    (qp,) = qps
    assert 1 == len(qp.qualifiers)
    (qual_obj,) = qp.qualifiers
    assert (
        qual_obj['policy_qualifier_id'].native
        == 'certification_practice_statement'
    )
    assert qual_obj['qualifier'].native == (
        'http://csrc.nist.gov/groups/ST/crypto_apps_infra/csor/'
        'pki_registration.html#PKITest'
    )



import datetime
import os

import pytest
from asn1crypto import cms, crl, ocsp, x509

from pyhanko_certvalidator import validate
from pyhanko_certvalidator.authority import CertTrustAnchor
from pyhanko_certvalidator.context import ACTargetDescription, ValidationContext
from pyhanko_certvalidator.errors import (
    InvalidCertificateError,
    PathValidationError,
    RevokedError,
)
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.revinfo.validate_crl import verify_crl
from pyhanko_certvalidator.revinfo.validate_ocsp import verify_ocsp_response

from .test_validate import FIXTURES_DIR

attr_cert_dir = os.path.join(FIXTURES_DIR, 'attribute-certs')
basic_aa_dir = os.path.join(attr_cert_dir, 'basic-aa')


def load_cert(fname) -> x509.Certificate:
    with open(fname, 'rb') as inf:
        return x509.Certificate.load(inf.read())


def load_attr_cert(fname) -> cms.AttributeCertificateV2:
    with open(fname, 'rb') as inf:
        return cms.AttributeCertificateV2.load(inf.read())


def load_crl(fname) -> crl.CertificateList:
    with open(fname, 'rb') as inf:
        return crl.CertificateList.load(inf.read())


def load_ocsp_response(fname) -> ocsp.OCSPResponse:
    with open(fname, 'rb') as inf:
        return ocsp.OCSPResponse.load(inf.read())


@pytest.mark.asyncio
async def test_basic_ac_validation_aacontrols_norev():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-norev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
    )

    result = await validate.async_validate_ac(ac, vc)
    assert len(result.aa_path) == 3
    assert 'role' in result.approved_attributes
    assert 'group' not in result.approved_attributes


@pytest.mark.asyncio
async def test_basic_ac_validation_bad_signature():
    ac = load_attr_cert(os.path.join(basic_aa_dir, 'aa', 'badsig.attr.crt'))

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
    )
    msg = 'signature could not be verified'
    with pytest.raises(InvalidCertificateError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_ac_validation_expired():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-norev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        moment=datetime.datetime(3000, 1, 1, tzinfo=datetime.timezone.utc),
    )
    msg = 'intermediate certificate 1 expired'
    with pytest.raises(PathValidationError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_basic_ac_validation_sig_algo_mismatch():
    ac = load_attr_cert(os.path.join(basic_aa_dir, 'aa', 'badsig.attr.crt'))
    # manipulate the signature algorithm
    ac = cms.AttributeCertificateV2(
        {
            'ac_info': ac['ac_info'],
            'signature_algorithm': {'algorithm': 'md5_rsa'},
            'signature': ac['signature'],
        }
    )
    ac.dump()

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
    )
    msg = 'algorithm declaration.*does not match'
    with pytest.raises(InvalidCertificateError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_basic_ac_validation_bad_aa_controls():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-norev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    # no AA controls on this one
    interm = load_cert(
        os.path.join(basic_aa_dir, 'root', 'interm-unrestricted.crt')
    )
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
    )

    msg = 'AA controls extension only present on part '
    with pytest.raises(PathValidationError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_basic_ac_validation_aa_controls_path_too_long():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-norev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    # no AA controls on this one
    interm = load_cert(
        os.path.join(basic_aa_dir, 'inbetween', 'interm-pathlen-violation.crt')
    )
    inbetween = load_cert(
        os.path.join(basic_aa_dir, 'root', 'inbetween-aa.crt')
    )
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa, inbetween],
    )

    msg = 'exceeds the maximum path length for an AA certificate'
    with pytest.raises(PathValidationError, match=msg):
        await validate.async_validate_ac(ac, vc)


def _load_targeted_ac():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-norev-targeted.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(
        os.path.join(basic_aa_dir, 'root', 'interm-unrestricted.crt')
    )
    aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'aa-unrestricted.crt'))
    return root, interm, aa, ac


@pytest.mark.asyncio
async def test_basic_ac_validation_no_targeting():
    root, interm, aa, ac = _load_targeted_ac()

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, aa],
    )

    msg = 'no targeting information'
    with pytest.raises(InvalidCertificateError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_basic_ac_validation_bad_targeting_name():
    root, interm, aa, ac = _load_targeted_ac()

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, aa],
        acceptable_ac_targets=ACTargetDescription(
            validator_names=[
                x509.GeneralName(
                    name='directory_name',
                    value=x509.Name.build(
                        {
                            'country_name': 'XX',
                            'organization_name': 'Testing Attribute Authority',
                            'organizational_unit_name': 'Validators',
                            'common_name': 'Not Validator',
                        }
                    ),
                )
            ]
        ),
    )

    msg = 'AC targeting'
    with pytest.raises(InvalidCertificateError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_basic_ac_validation_bad_targeting_group():
    root, interm, aa, ac = _load_targeted_ac()

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, aa],
        acceptable_ac_targets=ACTargetDescription(
            group_memberships=[
                x509.GeneralName(
                    name='directory_name',
                    value=x509.Name.build(
                        {
                            'country_name': 'XX',
                            'organization_name': 'Testing Attribute Authority',
                            'organizational_unit_name': 'Not Validators',
                        }
                    ),
                )
            ]
        ),
    )

    msg = 'AC targeting'
    with pytest.raises(InvalidCertificateError, match=msg):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_basic_ac_validation_good_targeting_name():
    root, interm, aa, ac = _load_targeted_ac()

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, aa],
        acceptable_ac_targets=ACTargetDescription(
            validator_names=[
                x509.GeneralName(
                    name='directory_name',
                    value=x509.Name.build(
                        {
                            'country_name': 'XX',
                            'organization_name': 'Testing Attribute Authority',
                            'organizational_unit_name': 'Validators',
                            'common_name': 'Validator',
                        }
                    ),
                )
            ]
        ),
    )

    result = await validate.async_validate_ac(ac, vc)
    assert len(result.aa_path) == 3
    assert 'role' in result.approved_attributes
    assert 'group' in result.approved_attributes


@pytest.mark.asyncio
async def test_basic_ac_validation_good_targeting_group():
    root, interm, aa, ac = _load_targeted_ac()

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, aa],
        acceptable_ac_targets=ACTargetDescription(
            group_memberships=[
                x509.GeneralName(
                    name='directory_name',
                    value=x509.Name.build(
                        {
                            'country_name': 'XX',
                            'organization_name': 'Testing Attribute Authority',
                            'organizational_unit_name': 'Validators',
                        }
                    ),
                )
            ]
        ),
    )

    result = await validate.async_validate_ac(ac, vc)
    assert len(result.aa_path) == 3
    assert 'role' in result.approved_attributes
    assert 'group' in result.approved_attributes


@pytest.mark.asyncio
async def test_match_holder_ac():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-norev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    alice = load_cert(os.path.join(basic_aa_dir, 'people-ca', 'alice.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
    )

    await validate.async_validate_ac(ac, vc, holder_cert=alice)


@pytest.mark.asyncio
async def test_match_holder_ac_mismatch():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-norev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    bob = load_cert(os.path.join(basic_aa_dir, 'people-ca', 'bob.crt'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
    )

    msg = 'Could not match.*base_certificate_id'
    with pytest.raises(InvalidCertificateError, match=msg):
        await validate.async_validate_ac(ac, vc, holder_cert=bob)


@pytest.mark.asyncio
async def test_ac_revoked():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    role_aa_crl = load_crl(
        os.path.join(basic_aa_dir, 'role-aa-some-revoked.crl')
    )

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        crls=[role_aa_crl],
        moment=datetime.datetime(
            year=2021, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )
    ac_path = ValidationPath(CertTrustAnchor(root), [interm, role_aa], ac)

    with pytest.raises(RevokedError):
        await verify_crl(ac, ac_path, vc)


@pytest.mark.asyncio
async def test_ac_unrevoked():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    role_aa_crl = load_crl(os.path.join(basic_aa_dir, 'role-aa-all-good.crl'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        crls=[role_aa_crl],
        moment=datetime.datetime(
            year=2019, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )
    ac_path = ValidationPath(CertTrustAnchor(root), [interm, role_aa], ac)

    await verify_crl(ac, ac_path, vc)


@pytest.mark.asyncio
async def test_ac_revoked_full_path_validation():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    role_aa_crl = load_crl(
        os.path.join(basic_aa_dir, 'role-aa-some-revoked.crl')
    )

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        crls=[role_aa_crl],
        moment=datetime.datetime(
            year=2021, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )

    with pytest.raises(RevokedError):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_ac_unrevoked_full_path_validation():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    role_aa_crl = load_crl(os.path.join(basic_aa_dir, 'role-aa-all-good.crl'))

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        crls=[role_aa_crl],
        moment=datetime.datetime(
            year=2019, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )
    await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_ac_revoked():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    ocsp_resp = load_ocsp_response(
        os.path.join(basic_aa_dir, 'alice-revoked.ors')
    )

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        ocsps=[ocsp_resp],
        moment=datetime.datetime(
            year=2021, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )
    ac_path = ValidationPath(CertTrustAnchor(root), [interm, role_aa], ac)

    with pytest.raises(RevokedError):
        await verify_ocsp_response(ac, ac_path, vc)


@pytest.mark.asyncio
async def test_ac_unrevoked():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    ocsp_resp = load_ocsp_response(
        os.path.join(basic_aa_dir, 'alice-all-good.ors')
    )

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        ocsps=[ocsp_resp],
        moment=datetime.datetime(
            year=2019, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )
    ac_path = ValidationPath(CertTrustAnchor(root), [interm, role_aa], ac)
    await verify_ocsp_response(ac, ac_path, vc)


@pytest.mark.asyncio
async def test_ac_revoked_full_path_validation():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    ocsp_resp = load_ocsp_response(
        os.path.join(basic_aa_dir, 'alice-revoked.ors')
    )

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        ocsps=[ocsp_resp],
        moment=datetime.datetime(
            year=2021, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )

    with pytest.raises(RevokedError):
        await validate.async_validate_ac(ac, vc)


@pytest.mark.asyncio
async def test_ac_unrevoked_full_path_validation():
    ac = load_attr_cert(
        os.path.join(basic_aa_dir, 'aa', 'alice-role-with-rev.attr.crt')
    )

    root = load_cert(os.path.join(basic_aa_dir, 'root', 'root.crt'))
    interm = load_cert(os.path.join(basic_aa_dir, 'root', 'interm-role.crt'))
    role_aa = load_cert(os.path.join(basic_aa_dir, 'interm', 'role-aa.crt'))

    ocsp_resp = load_ocsp_response(
        os.path.join(basic_aa_dir, 'alice-all-good.ors')
    )

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm, role_aa],
        ocsps=[ocsp_resp],
        moment=datetime.datetime(
            year=2019, month=12, day=12, tzinfo=datetime.timezone.utc
        ),
    )

    await validate.async_validate_ac(ac, vc)



from pyhanko_certvalidator.errors import ValidationError

__all__ = ['PastValidatePrecheckFailure', 'TimeSlideFailure']


class PastValidatePrecheckFailure(ValidationError):
    pass


class TimeSlideFailure(ValidationError):
    pass



import asyncio
from typing import Any, Dict, Union

import aiohttp

from ..api import DEFAULT_USER_AGENT
from ..common_utils import queue_fetch_task

__all__ = ['LazySession', 'AIOHttpMixin']


class LazySession:
    def __init__(self):
        self._session = None

    async def get_session(self):
        session = self._session
        if session is None:
            self._session = session = aiohttp.ClientSession()
        return session

    async def close(self):
        session = self._session
        if session is not None:
            await session.close()


class AIOHttpMixin:
    def __init__(
        self,
        session: Union[aiohttp.ClientSession, LazySession],
        user_agent=None,
        per_request_timeout=10,
    ):
        self._session = session
        self.user_agent = user_agent or DEFAULT_USER_AGENT
        self.per_request_timeout = per_request_timeout
        self.__results: Dict[Any, Any] = {}
        self.__result_events: Dict[Any, asyncio.Event] = {}
        super().__init__()

    async def get_session(self) -> aiohttp.ClientSession:
        session = self._session
        if isinstance(session, LazySession):
            return await session.get_session()
        else:
            return session

    def get_results(self):
        return {
            v for v in self.__results.values() if not isinstance(v, Exception)
        }

    def get_results_for_tag(self, tag):
        result = self.__results[tag]
        if isinstance(result, Exception):
            raise KeyError

    def _iter_results(self):
        for k, v in self.__results.items():
            if not isinstance(v, Exception):
                yield k, v

    async def _post_fetch_task(self, tag, async_fun):
        return await queue_fetch_task(
            self.__results, self.__result_events, tag, async_fun
        )



# coding: utf-8

from datetime import datetime

import pytest
from asn1crypto.util import timezone

from pyhanko_certvalidator import (
    CertificateValidator,
    PKIXValidationParams,
    ValidationContext,
)
from pyhanko_certvalidator.errors import (
    InvalidCertificateError,
    PathValidationError,
)
from tests.common import load_cert_object, load_nist_cert


@pytest.mark.asyncio
async def test_basic_certificate_validator_tls():
    cert = load_cert_object('mozilla.org.crt')
    other_certs = [load_cert_object('digicert-sha2-secure-server-ca.crt')]

    moment = datetime(2019, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    context = ValidationContext(moment=moment)
    validator = CertificateValidator(cert, other_certs, context)

    await validator.async_validate_tls('www.mozilla.org')


@pytest.mark.asyncio
async def test_basic_certificate_validator_tls_expired():
    cert = load_cert_object('mozilla.org.crt')
    other_certs = [load_cert_object('digicert-sha2-secure-server-ca.crt')]

    moment = datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    context = ValidationContext(moment=moment)
    validator = CertificateValidator(cert, other_certs, context)

    with pytest.raises(PathValidationError, match='expired'):
        await validator.async_validate_tls('www.mozilla.org')


@pytest.mark.asyncio
async def test_basic_certificate_validator_tls_invalid_hostname():
    cert = load_cert_object('mozilla.org.crt')
    other_certs = [load_cert_object('digicert-sha2-secure-server-ca.crt')]

    moment = datetime(2019, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    context = ValidationContext(moment=moment)
    validator = CertificateValidator(cert, other_certs, context)

    with pytest.raises(InvalidCertificateError, match='not valid'):
        await validator.async_validate_tls('google.com')


@pytest.mark.asyncio
async def test_basic_certificate_validator_tls_invalid_key_usage():
    cert = load_cert_object('mozilla.org.crt')
    other_certs = [load_cert_object('digicert-sha2-secure-server-ca.crt')]

    moment = datetime(2019, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    context = ValidationContext(moment=moment)
    validator = CertificateValidator(cert, other_certs, context)

    with pytest.raises(InvalidCertificateError, match='for the purpose'):
        await validator.async_validate_usage({'crl_sign'})


@pytest.mark.asyncio
async def test_basic_certificate_validator_tls_whitelist():
    cert = load_cert_object('mozilla.org.crt')
    other_certs = [load_cert_object('digicert-sha2-secure-server-ca.crt')]

    moment = datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    context = ValidationContext(
        whitelisted_certs=[cert.sha1_fingerprint], moment=moment
    )
    validator = CertificateValidator(cert, other_certs, context)

    # If whitelist does not work, this will raise exception for expiration
    await validator.async_validate_tls('www.mozilla.org')

    # If whitelist does not work, this will raise exception for hostname
    await validator.async_validate_tls('google.com')

    # If whitelist does not work, this will raise exception for key usage
    await validator.async_validate_usage({'crl_sign'})


@pytest.mark.asyncio
async def test_certvalidator_with_params():
    cert = load_nist_cert('ValidPolicyMappingTest12EE.crt')
    ca_certs = [load_nist_cert('TrustAnchorRootCertificate.crt')]
    other_certs = [load_nist_cert('P12Mapping1to3CACert.crt')]

    context = ValidationContext(
        trust_roots=ca_certs,
        other_certs=other_certs,
        revocation_mode="soft-fail",
        weak_hash_algos={'md2', 'md5'},
    )

    validator = CertificateValidator(
        cert,
        validation_context=context,
        pkix_params=PKIXValidationParams(
            user_initial_policy_set=frozenset(['2.16.840.1.101.3.2.1.48.1'])
        ),
    )
    path = await validator.async_validate_usage(key_usage={'digital_signature'})

    # check if we got the right policy processing
    # (i.e. if our params got through)
    qps = path.qualified_policies()

    (qp,) = qps
    assert 1 == len(qp.qualifiers)
    (qual_obj,) = qp.qualifiers
    assert qual_obj['policy_qualifier_id'].native == 'user_notice'
    assert qual_obj['qualifier']['explicit_text'].native == (
        'q7:  This is the user notice from qualifier 7 associated with '
        'NIST-test-policy-3.  This user notice should be displayed '
        'when  NIST-test-policy-1 is in the user-constrained-policy-set'
    )


@pytest.mark.asyncio
async def test_self_signed_with_policy():
    # tests whether a corner case in the policy validation logic when the
    # path length is zero is handled gracefully
    cert = load_cert_object('self-signed-with-policy.crt')
    context = ValidationContext(trust_roots=[cert], allow_fetching=False)
    validator = CertificateValidator(cert, validation_context=context)
    path = await validator.async_validate_usage({'digital_signature'})
    (qp,) = path.qualified_policies()
    # Note: the cert declares a concrete policy, but for the purposes
    # of PKIX validation, any policy is valid, since we're validating
    # a -signed certificate (so everything breaks down anyway)
    assert qp.user_domain_policy_id == 'any_policy'
    assert qp.issuer_domain_policy_id == 'any_policy'



import os
from datetime import datetime, timedelta, timezone

import pytest

from pyhanko_certvalidator import ValidationContext
from pyhanko_certvalidator.errors import PathValidationError, RevokedError
from pyhanko_certvalidator.policy_decl import (
    CertRevTrustPolicy,
    FreshnessReqType,
    RevocationCheckingPolicy,
)
from pyhanko_certvalidator.validate import async_validate_path

from .common import load_cert_object, load_crl, load_ocsp_response

freshness_dir = 'freshness'
certs = os.path.join('freshness', 'certs')


@pytest.mark.asyncio
async def test_cooldown_period_ok():
    req_policy = RevocationCheckingPolicy.from_legacy('require')
    policy = CertRevTrustPolicy(
        revocation_checking_policy=req_policy,
        freshness=timedelta(days=3),
        freshness_req_type=FreshnessReqType.TIME_AFTER_SIGNATURE,
    )
    root = load_cert_object(certs, 'root.crt')
    alice = load_cert_object(certs, 'alice.crt')
    interm = load_cert_object(certs, 'interm.crt')

    alice_ocsp = load_ocsp_response(freshness_dir, 'alice-2020-10-01.ors')
    root_crl = load_crl(freshness_dir, 'root-2020-10-01.crl')

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 10, 1, tzinfo=timezone.utc),
        best_signature_time=datetime(2020, 9, 18, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    await async_validate_path(vc, path)


@pytest.mark.asyncio
async def test_cooldown_period_too_early():
    req_policy = RevocationCheckingPolicy.from_legacy('require')
    policy = CertRevTrustPolicy(
        revocation_checking_policy=req_policy,
        freshness=timedelta(days=3),
        freshness_req_type=FreshnessReqType.TIME_AFTER_SIGNATURE,
    )
    root = load_cert_object(certs, 'root.crt')
    alice = load_cert_object(certs, 'alice.crt')
    interm = load_cert_object(certs, 'interm.crt')

    alice_ocsp = load_ocsp_response(freshness_dir, 'alice-2020-10-01.ors')
    root_crl = load_crl(freshness_dir, 'root-2020-10-01.crl')

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 10, 1, tzinfo=timezone.utc),
        best_signature_time=datetime(2020, 9, 30, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    with pytest.raises(PathValidationError, match='CRL.*recent enough'):
        await async_validate_path(vc, path)


@pytest.mark.asyncio
async def test_use_delta_ok():
    req_policy = RevocationCheckingPolicy.from_legacy('require')
    policy = CertRevTrustPolicy(
        revocation_checking_policy=req_policy,
        freshness=timedelta(days=9),
        freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
    )
    root = load_cert_object(certs, 'root.crt')
    alice = load_cert_object(certs, 'alice.crt')
    interm = load_cert_object(certs, 'interm.crt')

    alice_ocsp = load_ocsp_response(freshness_dir, 'alice-2020-10-01.ors')
    root_crl = load_crl(freshness_dir, 'root-2020-10-01.crl')

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 10, 1, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    await async_validate_path(vc, path)


@pytest.mark.asyncio
async def test_use_delta_stale():
    req_policy = RevocationCheckingPolicy.from_legacy('require')
    policy = CertRevTrustPolicy(
        revocation_checking_policy=req_policy,
        freshness=timedelta(hours=1),
        freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
    )
    root = load_cert_object(certs, 'root.crt')
    alice = load_cert_object(certs, 'alice.crt')
    interm = load_cert_object(certs, 'interm.crt')

    alice_ocsp = load_ocsp_response(freshness_dir, 'alice-2020-10-01.ors')
    root_crl = load_crl(freshness_dir, 'root-2020-10-01.crl')

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 10, 1, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    with pytest.raises(PathValidationError, match='CRL.*recent enough'):
        await async_validate_path(vc, path)


@pytest.mark.asyncio
async def test_use_most_recent():
    req_policy = RevocationCheckingPolicy.from_legacy('require')
    policy = CertRevTrustPolicy(
        revocation_checking_policy=req_policy,
        freshness=timedelta(days=20),  # some ridiculous value
        freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
    )
    root = load_cert_object(certs, 'root.crt')
    alice = load_cert_object(certs, 'alice.crt')
    interm = load_cert_object(certs, 'interm.crt')

    alice_ocsp_older = load_ocsp_response(freshness_dir, 'alice-2020-11-29.ors')
    alice_ocsp_recent = load_ocsp_response(
        freshness_dir, 'alice-2020-12-10.ors'
    )
    root_crl = load_crl(freshness_dir, 'root-2020-12-10.crl')

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp_older, alice_ocsp_recent],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 12, 10, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    with pytest.raises(RevokedError):
        await async_validate_path(vc, path)

    # Double-check: the validator should be fooled if we don't include the
    #  second OCSP response because of the very lenient time delta allowed
    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp_older],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 12, 10, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    await async_validate_path(vc, path)


@pytest.mark.asyncio
async def test_discard_post_validation_time():
    req_policy = RevocationCheckingPolicy.from_legacy('require')
    policy = CertRevTrustPolicy(
        revocation_checking_policy=req_policy,
        freshness=timedelta(days=20),  # some ridiculous value
        freshness_req_type=FreshnessReqType.MAX_DIFF_REVOCATION_VALIDATION,
    )
    root = load_cert_object(certs, 'root.crt')
    alice = load_cert_object(certs, 'alice.crt')
    interm = load_cert_object(certs, 'interm.crt')

    alice_ocsp_older = load_ocsp_response(freshness_dir, 'alice-2020-11-29.ors')
    alice_ocsp_recent = load_ocsp_response(
        freshness_dir, 'alice-2020-12-10.ors'
    )
    root_crl = load_crl(freshness_dir, 'root-2020-11-29.crl')

    vc = ValidationContext(
        trust_roots=[root],
        other_certs=[interm],
        ocsps=[alice_ocsp_older, alice_ocsp_recent],
        crls=[root_crl],
        revinfo_policy=policy,
        moment=datetime(2020, 11, 29, tzinfo=timezone.utc),
    )
    (path,) = await vc.path_builder.async_build_paths(alice)
    await async_validate_path(vc, path)



import logging
from typing import Iterable, Union

import aiohttp
from asn1crypto import cms, x509

from ...errors import CertificateFetchError
from ..api import CertificateFetcher
from ..common_utils import (
    ACCEPTABLE_CERT_DER_ALIASES,
    ACCEPTABLE_CERT_PEM_ALIASES,
    ACCEPTABLE_PKCS7_DER_ALIASES,
    ACCEPTABLE_STRICT_CERT_CONTENT_TYPES,
    complete_certificate_fetch_jobs,
    gather_aia_issuer_urls,
    unpack_cert_content,
)
from .util import AIOHttpMixin, LazySession

logger = logging.getLogger(__name__)


class AIOHttpCertificateFetcher(CertificateFetcher, AIOHttpMixin):
    def __init__(
        self,
        session: Union[aiohttp.ClientSession, LazySession],
        user_agent=None,
        per_request_timeout=10,
        permit_pem=True,
    ):
        super().__init__(session, user_agent, per_request_timeout)
        self.permit_pem = permit_pem

    async def fetch_certs(self, url, url_origin_type):
        """
        Fetch one or more certificates from a URL.

        :param url:
            URL to fetch.
        :param url_origin_type:
            Parameter indicating where the URL came from (e.g. 'CRL'),
            for error reporting purposes.
        :raises:
            CertificateFetchError - when a network I/O or decoding error occurs
        :return:
            An iterable of asn1crypto.x509.Certificate objects.
        """

        async def task():
            try:
                logger.info(f"Fetching certificates from {url}...")
                return await _grab_certs(
                    url,
                    permit_pem=self.permit_pem,
                    timeout=self.per_request_timeout,
                    user_agent=self.user_agent,
                    session=await self.get_session(),
                    url_origin_type=url_origin_type,
                )
            except (ValueError, aiohttp.ClientError) as e:
                msg = f"Failed to fetch certificate(s) from url {url}."
                logger.debug(msg, exc_info=e)
                raise CertificateFetchError(msg)

        return await self._post_fetch_task(url, task)

    def fetch_cert_issuers(
        self, cert: Union[x509.Certificate, cms.AttributeCertificateV2]
    ):
        fetch_jobs = [
            self.fetch_certs(url, url_origin_type='certificate')
            for url in gather_aia_issuer_urls(cert)
        ]

        if isinstance(cert, x509.Certificate):
            target = cert.subject.human_friendly
        else:
            # TODO log audit ID
            target = "attribute certificate"
        logger.info(f"Retrieving issuer certs for {target}...")
        return complete_certificate_fetch_jobs(fetch_jobs)

    def fetch_crl_issuers(self, certificate_list):
        fetch_jobs = [
            self.fetch_certs(url, url_origin_type='CRL')
            for url in certificate_list.issuer_cert_urls
        ]
        return complete_certificate_fetch_jobs(fetch_jobs)

    def fetched_certs(self) -> Iterable[x509.Certificate]:
        return self.get_results()


async def _grab_certs(
    url,
    *,
    user_agent,
    session: aiohttp.ClientSession,
    url_origin_type,
    timeout,
    permit_pem=True,
):
    """
    Grab one or more certificates from a caIssuers URL.

    We accept two types of content in the response:
      - A single DER-encoded X.509 certificate
      - A PKCS#7 'certs-only' SignedData message
      - PEM-encoded certificates (if permit_pem=True)

    Note: strictly speaking, you're not supposed to use PEM to serve certs for
    AIA purposes in PEM format, but people do it anyway.
    """

    if permit_pem:
        acceptable_cts = (
            ACCEPTABLE_STRICT_CERT_CONTENT_TYPES
            | ACCEPTABLE_CERT_PEM_ALIASES
            | ACCEPTABLE_CERT_DER_ALIASES
            | ACCEPTABLE_PKCS7_DER_ALIASES
        )
    else:
        acceptable_cts = ACCEPTABLE_STRICT_CERT_CONTENT_TYPES

    headers = {'Accept': ','.join(acceptable_cts), 'User-Agent': user_agent}
    cl_timeout = aiohttp.ClientTimeout(timeout)
    async with session.get(
        url=url, headers=headers, timeout=cl_timeout, raise_for_status=True
    ) as response:
        response_data = await response.read()
        try:
            content_type = response.headers['Content-Type'].strip()
            if content_type not in acceptable_cts:
                ct_err = (
                    f"Unacceptable content type '{repr(content_type)}' "
                    f"when fetching issuer certificate for {url_origin_type} "
                    f"from URL {url}."
                )
                raise aiohttp.ContentTypeError(
                    response.request_info,
                    response.history,
                    message=ct_err,
                    headers=response.headers,
                )
        except KeyError:
            content_type = None
    certs = unpack_cert_content(response_data, content_type, url, permit_pem)
    return list(certs)



import logging
from typing import Dict, Iterable, List, Union

import aiohttp
from asn1crypto import cms, crl, pem, x509

from ... import errors
from ...util import get_relevant_crl_dps, issuer_serial
from ..api import CRLFetcher
from ..common_utils import crl_job_results_as_completed
from .util import AIOHttpMixin, LazySession

logger = logging.getLogger(__name__)


class AIOHttpCRLFetcher(CRLFetcher, AIOHttpMixin):
    def __init__(
        self,
        session: Union[aiohttp.ClientSession, LazySession],
        user_agent=None,
        per_request_timeout=10,
    ):
        super().__init__(session, user_agent, per_request_timeout)
        self._by_cert: Dict[bytes, List[crl.CertificateList]] = {}

    async def fetch(
        self,
        cert: Union[x509.Certificate, cms.AttributeCertificateV2],
        *,
        use_deltas=True,
    ):
        iss_serial = issuer_serial(cert)
        try:
            return self._by_cert[iss_serial]
        except KeyError:
            pass

        results = []
        async for fetched_crl in self._fetch(cert, use_deltas=use_deltas):
            results.append(fetched_crl)
        self._by_cert[iss_serial] = results
        return results

    async def _fetch(self, cert: x509.Certificate, *, use_deltas):
        sources = get_relevant_crl_dps(cert, use_deltas=use_deltas)

        if not sources:
            return

        if isinstance(cert, x509.Certificate):
            target = cert.subject.human_friendly
        else:
            # TODO log audit ID
            target = "attribute certificate"
        logger.info(f"Retrieving CRLs for {target}...")

        def _fetch_jobs():
            for distribution_point in sources:
                url = distribution_point.url
                # Only fetch CRLs over http
                #  (or https, but that doesn't really happen all that often)
                # In particular, don't attempt to grab CRLs over LDAP
                if url.startswith('http'):
                    yield self._single_fetch(url)

        # when the issue with .crl_distribution_points is fixed,
        # we should handle at_least_one_success and last_e on a per-DP basis
        async for result in crl_job_results_as_completed(_fetch_jobs()):
            yield result

    async def _single_fetch(self, url):
        async def task():
            return await _grab_crl(
                url,
                user_agent=self.user_agent,
                session=await self.get_session(),
                timeout=self.per_request_timeout,
            )

        return await self._post_fetch_task(url, task)

    def fetched_crls(self) -> Iterable[crl.CertificateList]:
        return {crl_ for crl_ in self.get_results()}

    def fetched_crls_for_cert(self, cert) -> Iterable[crl.CertificateList]:
        return self._by_cert[issuer_serial(cert)]


async def _grab_crl(
    url, *, user_agent, session: aiohttp.ClientSession, timeout
):
    """
    Fetches a CRL and parses it

    :param url:
        A unicode string of the URL to fetch the CRL from

    :param user_agent:
        A unicode string of the user agent to use when fetching the URL

    :param session:
        ``aiohttp`` client session to use.

    :param timeout:
        Timeout in seconds.

    :return:
        An asn1crypto.crl.CertificateList object
    """
    try:
        logger.info(f"Requesting CRL from {url}...")
        headers = {'Accept': 'application/pkix-crl', 'User-Agent': user_agent}
        cl_timeout = aiohttp.ClientTimeout(total=timeout)
        async with session.get(
            url=url, headers=headers, timeout=cl_timeout, raise_for_status=True
        ) as response:
            data = await response.read()
        if pem.detect(data):
            _, _, data = pem.unarmor(data)
        return crl.CertificateList.load(data)
    except (ValueError, aiohttp.ClientError) as e:
        raise errors.CRLFetchError(
            f"Failure to fetch CRL from URL {url}"
        ) from e



import os

import pytest
from asn1crypto import x509

from pyhanko_certvalidator.authority import (
    CertTrustAnchor,
    NamedKeyAuthority,
    TrustAnchor,
    TrustQualifiers,
)
from pyhanko_certvalidator.context import ValidationContext
from pyhanko_certvalidator.errors import PathValidationError
from pyhanko_certvalidator.name_trees import (
    GeneralNameType,
    x509_names_to_subtrees,
)
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.policy_decl import PKIXValidationParams
from pyhanko_certvalidator.validate import async_validate_path
from tests.common import load_nist_cert


def test_extract_policy():
    # I know this isn't a CA cert, but it's a convenient one to use
    crt = load_nist_cert('ValidCertificatePathTest1EE.crt')
    anchor = CertTrustAnchor(crt, derive_default_quals_from_cert=True)
    params = anchor.trust_qualifiers.standard_parameters
    nist_test_policy = '2.16.840.1.101.3.2.1.48.1'
    assert params.user_initial_policy_set == {nist_test_policy}


def test_extract_permitted_subtrees():
    crt = load_nist_cert('nameConstraintsDN1CACert.crt')
    anchor = CertTrustAnchor(crt, derive_default_quals_from_cert=True)
    params = anchor.trust_qualifiers.standard_parameters
    dirname_trs = params.initial_permitted_subtrees[
        GeneralNameType.DIRECTORY_NAME
    ]
    assert len(dirname_trs) == 1
    (tree,) = dirname_trs
    expected_name = x509.Name.build(
        {
            'organizational_unit_name': 'permittedSubtree1',
            'organization_name': 'Test Certificates 2011',
            'country_name': 'US',
        }
    )
    assert tree.tree_base.value == expected_name


@pytest.mark.asyncio
async def test_validate_with_derived():
    crt = load_nist_cert('nameConstraintsDN1CACert.crt')
    anchor = CertTrustAnchor(crt, derive_default_quals_from_cert=True)
    ee = load_nist_cert('InvalidDNnameConstraintsTest2EE.crt')
    context = ValidationContext(
        trust_roots=[anchor],
        revocation_mode='soft-fail',
    )
    (path,) = await context.path_builder.async_build_paths(ee)
    assert path.pkix_len == 1
    with pytest.raises(PathValidationError, match='not all names.*permitted'):
        await async_validate_path(context, path)


@pytest.mark.asyncio
async def test_validate_with_merged_permitted_subtrees():
    crt = load_nist_cert('nameConstraintsDN1CACert.crt')
    anchor = CertTrustAnchor(crt, derive_default_quals_from_cert=True)
    ee = load_nist_cert('ValidDNnameConstraintsTest1EE.crt')
    context = ValidationContext(
        trust_roots=[anchor],
        revocation_mode='soft-fail',
    )
    (path,) = await context.path_builder.async_build_paths(ee)
    assert path.pkix_len == 1

    # this should be OK
    await async_validate_path(context, path)
    # merge in an extra name constraint
    extra_name = x509.Name.build(
        {
            'organizational_unit_name': 'someNameYouDontHave',
            'organization_name': 'Test Certificates 2011',
            'country_name': 'US',
        }
    )
    extra_params = PKIXValidationParams(
        initial_permitted_subtrees=x509_names_to_subtrees([extra_name])
    )
    with pytest.raises(PathValidationError, match='not all names.*permitted'):
        await async_validate_path(context, path, parameters=extra_params)


@pytest.mark.asyncio
async def test_validate_with_merged_excluded_subtrees():
    crt = load_nist_cert('nameConstraintsDN3CACert.crt')
    anchor = CertTrustAnchor(crt, derive_default_quals_from_cert=True)
    ee = load_nist_cert('ValidDNnameConstraintsTest6EE.crt')
    context = ValidationContext(
        trust_roots=[anchor],
        revocation_mode='soft-fail',
    )
    (path,) = await context.path_builder.async_build_paths(ee)
    assert path.pkix_len == 1

    # this should be OK
    await async_validate_path(context, path)
    # merge in an extra name constraint
    extra_name = x509.Name.build(
        {
            'organizational_unit_name': 'permittedSubtree1',
            'organization_name': 'Test Certificates 2011',
            'country_name': 'US',
        }
    )
    extra_params = PKIXValidationParams(
        initial_excluded_subtrees=x509_names_to_subtrees([extra_name])
    )
    with pytest.raises(PathValidationError, match='some names.*excluded'):
        await async_validate_path(context, path, parameters=extra_params)


@pytest.mark.asyncio
async def test_validate_with_certless_root():
    crt = load_nist_cert('nameConstraintsDN1CACert.crt')
    # manually build params
    permitted = x509.Name.build(
        {
            'organizational_unit_name': 'permittedSubtree1',
            'organization_name': 'Test Certificates 2011',
            'country_name': 'US',
        }
    )
    extra_params = PKIXValidationParams(
        initial_permitted_subtrees=x509_names_to_subtrees([permitted])
    )
    anchor = TrustAnchor(
        NamedKeyAuthority(crt.subject, crt.public_key),
        quals=TrustQualifiers(standard_parameters=extra_params),
    )
    ee = load_nist_cert('ValidDNnameConstraintsTest1EE.crt')
    context = ValidationContext(
        trust_roots=[anchor],
        revocation_mode='soft-fail',
    )
    (path,) = await context.path_builder.async_build_paths(ee)
    assert path.pkix_len == 1

    assert isinstance(path.first, x509.Certificate)
    assert path.trust_anchor is anchor

    await async_validate_path(context, path, parameters=extra_params)


@pytest.mark.asyncio
async def test_validate_with_certless_root_failure():
    crt = load_nist_cert('nameConstraintsDN1CACert.crt')
    # manually build params
    permitted = x509.Name.build(
        {
            'organizational_unit_name': 'someNameYouDontHave',
            'organization_name': 'Test Certificates 2011',
            'country_name': 'US',
        }
    )
    extra_params = PKIXValidationParams(
        initial_permitted_subtrees=x509_names_to_subtrees([permitted])
    )
    anchor = TrustAnchor(
        NamedKeyAuthority(crt.subject, crt.public_key),
        quals=TrustQualifiers(standard_parameters=extra_params),
    )
    ee = load_nist_cert('ValidDNnameConstraintsTest1EE.crt')
    context = ValidationContext(
        trust_roots=[anchor],
        revocation_mode='soft-fail',
    )
    (path,) = await context.path_builder.async_build_paths(ee)
    assert path.pkix_len == 1

    assert isinstance(path.first, x509.Certificate)
    assert path.trust_anchor is anchor
    with pytest.raises(PathValidationError, match='not all names.*permitted'):
        await async_validate_path(context, path, parameters=extra_params)


@pytest.mark.asyncio
async def test_validate_empty_path_certless_root():
    crt = load_nist_cert('nameConstraintsDN1CACert.crt')
    anchor = TrustAnchor(
        NamedKeyAuthority(crt.subject, crt.public_key),
    )
    context = ValidationContext(
        trust_roots=[anchor],
        revocation_mode='soft-fail',
    )

    trivial_path = ValidationPath(trust_anchor=anchor, interm=[], leaf=None)
    await async_validate_path(context, trivial_path)



# coding: utf-8

import pytest

from pyhanko_certvalidator.context import ValidationContext
from pyhanko_certvalidator.fetchers import aiohttp_fetchers, requests_fetchers
from pyhanko_certvalidator.revinfo.validate_crl import verify_crl

from .common import load_cert_object
from .constants import TEST_REQUEST_TIMEOUT


async def _test_with_fetchers(fetchers):
    intermediate = load_cert_object('digicert-g5-ecc-sha384-2021-ca1.crt')
    root = load_cert_object('digicert-root-g5.crt')

    crls = await fetchers.crl_fetcher.fetch(intermediate)
    context = ValidationContext(
        trust_roots=[root], crls=crls, fetchers=fetchers
    )
    paths = await context.path_builder.async_build_paths(intermediate)
    path = paths[0]

    await verify_crl(intermediate, path, context)


@pytest.mark.asyncio
async def test_fetch_crl_aiohttp():
    fb = aiohttp_fetchers.AIOHttpFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    )
    async with fb as fetchers:
        await _test_with_fetchers(fetchers)


@pytest.mark.asyncio
async def test_fetch_requests():
    fetchers = requests_fetchers.RequestsFetcherBackend(
        per_request_timeout=TEST_REQUEST_TIMEOUT
    ).get_fetchers()
    await _test_with_fetchers(fetchers)



from typing import Optional

import aiohttp

from ..api import FetcherBackend, Fetchers
from .cert_fetch_client import AIOHttpCertificateFetcher
from .crl_client import AIOHttpCRLFetcher
from .ocsp_client import AIOHttpOCSPFetcher
from .util import LazySession

__all__ = ['AIOHttpFetcherBackend']


class AIOHttpFetcherBackend(FetcherBackend):
    def __init__(
        self,
        session: Optional[aiohttp.ClientSession] = None,
        per_request_timeout=10,
    ):
        self.session = session or LazySession()
        self.per_request_timeout = per_request_timeout

    def get_fetchers(self) -> Fetchers:
        session = self.session
        to = self.per_request_timeout
        return Fetchers(
            ocsp_fetcher=AIOHttpOCSPFetcher(session, per_request_timeout=to),
            crl_fetcher=AIOHttpCRLFetcher(session, per_request_timeout=to),
            cert_fetcher=AIOHttpCertificateFetcher(
                session, per_request_timeout=to
            ),
        )

    async def close(self):
        session = self.session
        # only close the session if it's a lazy session;
        # a session passed in by the caller is their own responsibility
        if isinstance(session, LazySession):
            await session.close()
